# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
#   quantization operations
#   pruning operations
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""Utility functions for building models."""
from __future__ import print_function

import collections
import os
import time
import numpy as np
import six
import tensorflow as tf

from tensorflow.python.ops import lookup_ops
from nlp_architect.models.gnmt.utils import iterator_utils
from nlp_architect.models.gnmt.utils import misc_utils as utils
from nlp_architect.models.gnmt.utils import vocab_utils

__all__ = [
    "get_initializer", "get_device_str", "create_train_model",
    "create_eval_model", "create_infer_model",
    "create_emb_for_encoder_and_decoder", "create_rnn_cell", "gradient_clip",
    "create_or_load_model", "load_model", "avg_checkpoints",
    "compute_perplexity"
]

# If a vocab size is greater than this value, put the embedding on cpu instead
VOCAB_SIZE_THRESHOLD_CPU = 50000

# Collection for all the tensors involved in the quantization process
_QUANTIZATION_COLLECTION = 'qunatization'


def get_initializer(init_op, seed=None, init_weight=None):
    """Create an initializer. init_weight is only for uniform."""
    if init_op == "uniform":
        assert init_weight
        return tf.random_uniform_initializer(
            -init_weight, init_weight, seed=seed)
    elif init_op == "glorot_normal":
        return tf.keras.initializers.glorot_normal(
            seed=seed)
    elif init_op == "glorot_uniform":
        return tf.keras.initializers.glorot_uniform(
            seed=seed)
    else:
        raise ValueError("Unknown init_op %s" % init_op)


def get_device_str(device_id, num_gpus):
    """Return a device string for multi-GPU setup."""
    if num_gpus == 0:
        return "/cpu:0"
    device_str_output = "/gpu:%d" % (device_id % num_gpus)
    return device_str_output


class ExtraArgs(collections.namedtuple(
    "ExtraArgs", ("single_cell_fn", "model_device_fn",
                  "attention_mechanism_fn", "encoder_emb_lookup_fn"))):
    pass


class TrainModel(
    collections.namedtuple("TrainModel", ("graph", "model", "iterator",
                                          "skip_count_placeholder"))):
    pass


def create_train_model(
        model_creator, hparams, scope=None, num_workers=1, jobid=0,
        extra_args=None):
    """Create train graph, model, and iterator."""
    src_file = "%s.%s" % (hparams.train_prefix, hparams.src)
    tgt_file = "%s.%s" % (hparams.train_prefix, hparams.tgt)
    src_vocab_file = hparams.src_vocab_file
    tgt_vocab_file = hparams.tgt_vocab_file

    graph = tf.Graph()

    with graph.as_default(), tf.container(scope or "train"):
        src_vocab_table, tgt_vocab_table = vocab_utils.create_vocab_tables(
            src_vocab_file, tgt_vocab_file, hparams.share_vocab)

        src_dataset = tf.data.TextLineDataset(tf.gfile.Glob(src_file))
        tgt_dataset = tf.data.TextLineDataset(tf.gfile.Glob(tgt_file))
        skip_count_placeholder = tf.placeholder(shape=(), dtype=tf.int64)

        iterator = iterator_utils.get_iterator(
            src_dataset,
            tgt_dataset,
            src_vocab_table,
            tgt_vocab_table,
            batch_size=hparams.batch_size,
            sos=hparams.sos,
            eos=hparams.eos,
            random_seed=hparams.random_seed,
            num_buckets=hparams.num_buckets,
            src_max_len=hparams.src_max_len,
            tgt_max_len=hparams.tgt_max_len,
            skip_count=skip_count_placeholder,
            num_shards=num_workers,
            shard_index=jobid,
            use_char_encode=hparams.use_char_encode)

        # Note: One can set model_device_fn to
        # `tf.train.replica_device_setter(ps_tasks)` for distributed training.
        model_device_fn = None
        if extra_args:
            model_device_fn = extra_args.model_device_fn
        with tf.device(model_device_fn):
            model = model_creator(
                hparams,
                iterator=iterator,
                mode=tf.contrib.learn.ModeKeys.TRAIN,
                source_vocab_table=src_vocab_table,
                target_vocab_table=tgt_vocab_table,
                scope=scope,
                extra_args=extra_args)

    return TrainModel(
        graph=graph,
        model=model,
        iterator=iterator,
        skip_count_placeholder=skip_count_placeholder)


class EvalModel(
    collections.namedtuple("EvalModel",
                           ("graph", "model", "src_file_placeholder",
                            "tgt_file_placeholder", "iterator"))):
    pass


def create_eval_model(model_creator, hparams, scope=None, extra_args=None):
    """Create train graph, model, src/tgt file holders, and iterator."""
    src_vocab_file = hparams.src_vocab_file
    tgt_vocab_file = hparams.tgt_vocab_file
    graph = tf.Graph()

    with graph.as_default(), tf.container(scope or "eval"):
        src_vocab_table, tgt_vocab_table = vocab_utils.create_vocab_tables(
            src_vocab_file, tgt_vocab_file, hparams.share_vocab)
        reverse_tgt_vocab_table = lookup_ops.index_to_string_table_from_file(
            tgt_vocab_file, default_value=vocab_utils.UNK)

        src_file_placeholder = tf.placeholder(shape=(), dtype=tf.string)
        tgt_file_placeholder = tf.placeholder(shape=(), dtype=tf.string)
        src_dataset = tf.data.TextLineDataset(src_file_placeholder)
        tgt_dataset = tf.data.TextLineDataset(tgt_file_placeholder)
        iterator = iterator_utils.get_iterator(
            src_dataset,
            tgt_dataset,
            src_vocab_table,
            tgt_vocab_table,
            hparams.batch_size,
            sos=hparams.sos,
            eos=hparams.eos,
            random_seed=hparams.random_seed,
            num_buckets=hparams.num_buckets,
            src_max_len=hparams.src_max_len_infer,
            tgt_max_len=hparams.tgt_max_len_infer,
            use_char_encode=hparams.use_char_encode)
        model = model_creator(
            hparams,
            iterator=iterator,
            mode=tf.contrib.learn.ModeKeys.EVAL,
            source_vocab_table=src_vocab_table,
            target_vocab_table=tgt_vocab_table,
            reverse_target_vocab_table=reverse_tgt_vocab_table,
            scope=scope,
            extra_args=extra_args)
    return EvalModel(
        graph=graph,
        model=model,
        src_file_placeholder=src_file_placeholder,
        tgt_file_placeholder=tgt_file_placeholder,
        iterator=iterator)


class InferModel(
    collections.namedtuple("InferModel",
                           ("graph", "model", "src_placeholder",
                            "batch_size_placeholder", "iterator"))):
    pass


def create_infer_model(model_creator, hparams, scope=None, extra_args=None):
    """Create inference model."""
    graph = tf.Graph()
    src_vocab_file = hparams.src_vocab_file
    tgt_vocab_file = hparams.tgt_vocab_file

    with graph.as_default(), tf.container(scope or "infer"):
        src_vocab_table, tgt_vocab_table = vocab_utils.create_vocab_tables(
            src_vocab_file, tgt_vocab_file, hparams.share_vocab)
        reverse_tgt_vocab_table = lookup_ops.index_to_string_table_from_file(
            tgt_vocab_file, default_value=vocab_utils.UNK)

        src_placeholder = tf.placeholder(shape=[None], dtype=tf.string)
        batch_size_placeholder = tf.placeholder(shape=[], dtype=tf.int64)

        src_dataset = tf.data.Dataset.from_tensor_slices(
            src_placeholder)
        iterator = iterator_utils.get_infer_iterator(
            src_dataset,
            src_vocab_table,
            batch_size=batch_size_placeholder,
            eos=hparams.eos,
            src_max_len=hparams.src_max_len_infer,
            use_char_encode=hparams.use_char_encode)
        model = model_creator(
            hparams,
            iterator=iterator,
            mode=tf.contrib.learn.ModeKeys.INFER,
            source_vocab_table=src_vocab_table,
            target_vocab_table=tgt_vocab_table,
            reverse_target_vocab_table=reverse_tgt_vocab_table,
            scope=scope,
            extra_args=extra_args)
    return InferModel(
        graph=graph,
        model=model,
        src_placeholder=src_placeholder,
        batch_size_placeholder=batch_size_placeholder,
        iterator=iterator)


def _get_embed_device(vocab_size):
    """Decide on which device to place an embed matrix given its vocab size."""
    if vocab_size > VOCAB_SIZE_THRESHOLD_CPU:
        return "/cpu:0"
    else:
        return "/gpu:0"


def _create_pretrained_emb_from_txt(
        vocab_file, embed_file, num_trainable_tokens=3, dtype=tf.float32,
        scope=None):
    """Load pretrain embeding from embed_file, and return an embedding matrix.

    Args:
      embed_file: Path to a Glove formated embedding txt file.
      num_trainable_tokens: Make the first n tokens in the vocab file as trainable
        variables. Default is 3, which is "<unk>", "<s>" and "</s>".
    """
    vocab, _ = vocab_utils.load_vocab(vocab_file)
    trainable_tokens = vocab[:num_trainable_tokens]

    utils.print_out("# Using pretrained embedding: %s." % embed_file)
    utils.print_out("  with trainable tokens: ")

    emb_dict, emb_size = vocab_utils.load_embed_txt(embed_file)
    for token in trainable_tokens:
        utils.print_out("    %s" % token)
        if token not in emb_dict:
            emb_dict[token] = [0.0] * emb_size

    emb_mat = np.array(
        [emb_dict[token] for token in vocab], dtype=dtype.as_numpy_dtype())
    emb_mat = tf.constant(emb_mat)
    emb_mat_const = tf.slice(emb_mat, [num_trainable_tokens, 0], [-1, -1])
    with tf.variable_scope(scope or "pretrain_embeddings", dtype=dtype) as scope:
        with tf.device(_get_embed_device(num_trainable_tokens)):
            emb_mat_var = tf.get_variable(
                "emb_mat_var", [num_trainable_tokens, emb_size])
    return tf.concat([emb_mat_var, emb_mat_const], 0)


def _create_or_load_embed(embed_name, vocab_file, embed_file,
                          vocab_size, embed_size, dtype, embed_type='dense'):
    """Create a new or load an existing embedding matrix."""
    if vocab_file and embed_file:
        embedding = _create_pretrained_emb_from_txt(vocab_file, embed_file)
    else:
        with tf.device(_get_embed_device(vocab_size)):
            if embed_type == 'dense':
                embedding = tf.get_variable(
                    embed_name, [vocab_size, embed_size], dtype)
            elif embed_type == 'sparse':
                embedding = tf.get_variable(
                    embed_name, [vocab_size, embed_size], dtype)
                embedding = tf.contrib.model_pruning.apply_mask(embedding, embed_name)
            else:
                raise ValueError("Unknown embedding type %s!" % embed_type)
    return embedding


def create_emb_for_encoder_and_decoder(share_vocab,
                                       src_vocab_size,
                                       tgt_vocab_size,
                                       src_embed_size,
                                       tgt_embed_size,
                                       embed_type='dense',
                                       dtype=tf.float32,
                                       num_enc_partitions=0,
                                       num_dec_partitions=0,
                                       src_vocab_file=None,
                                       tgt_vocab_file=None,
                                       src_embed_file=None,
                                       tgt_embed_file=None,
                                       use_char_encode=False,
                                       scope=None):
    """Create embedding matrix for both encoder and decoder.

    Args:
      share_vocab: A boolean. Whether to share embedding matrix for both
        encoder and decoder.
      src_vocab_size: An integer. The source vocab size.
      tgt_vocab_size: An integer. The target vocab size.
      src_embed_size: An integer. The embedding dimension for the encoder's
        embedding.
      tgt_embed_size: An integer. The embedding dimension for the decoder's
        embedding.
      dtype: dtype of the embedding matrix. Default to float32.
      num_enc_partitions: number of partitions used for the encoder's embedding
        vars.
      num_dec_partitions: number of partitions used for the decoder's embedding
        vars.
      scope: VariableScope for the created subgraph. Default to "embedding".

    Returns:
      embedding_encoder: Encoder's embedding matrix.
      embedding_decoder: Decoder's embedding matrix.

    Raises:
      ValueError: if use share_vocab but source and target have different vocab
        size.
    """
    if num_enc_partitions <= 1:
        enc_partitioner = None
    else:
        # Note: num_partitions > 1 is required for distributed training due to
        # embedding_lookup tries to colocate single partition-ed embedding variable
        # with lookup ops. This may cause embedding variables being placed on worker
        # jobs.
        enc_partitioner = tf.fixed_size_partitioner(num_enc_partitions)

    if num_dec_partitions <= 1:
        dec_partitioner = None
    else:
        # Note: num_partitions > 1 is required for distributed training due to
        # embedding_lookup tries to colocate single partition-ed embedding variable
        # with lookup ops. This may cause embedding variables being placed on worker
        # jobs.
        dec_partitioner = tf.fixed_size_partitioner(num_dec_partitions)

    if src_embed_file and enc_partitioner:
        raise ValueError(
            "Can't set num_enc_partitions > 1 when using pretrained encoder "
            "embedding")

    if tgt_embed_file and dec_partitioner:
        raise ValueError(
            "Can't set num_dec_partitions > 1 when using pretrained decdoer "
            "embedding")

    with tf.variable_scope(scope or "embeddings", dtype=dtype, partitioner=enc_partitioner):
        # Share embedding
        if share_vocab:
            if src_vocab_size != tgt_vocab_size:
                raise ValueError("Share embedding but different src/tgt vocab sizes"
                                 " %d vs. %d" % (src_vocab_size, tgt_vocab_size))
            assert src_embed_size == tgt_embed_size
            utils.print_out("# Use the same embedding for source and target")
            vocab_file = src_vocab_file or tgt_vocab_file
            embed_file = src_embed_file or tgt_embed_file

            embedding_encoder = _create_or_load_embed(
                "embedding_share", vocab_file, embed_file,
                src_vocab_size, src_embed_size, dtype, embed_type=embed_type)
            embedding_decoder = embedding_encoder
        else:
            if not use_char_encode:
                with tf.variable_scope("encoder", partitioner=enc_partitioner):
                    embedding_encoder = _create_or_load_embed(
                        "embedding_encoder", src_vocab_file, src_embed_file,
                        src_vocab_size, src_embed_size, dtype, embed_type=embed_type)
            else:
                embedding_encoder = None

            with tf.variable_scope("decoder", partitioner=dec_partitioner):
                embedding_decoder = _create_or_load_embed(
                    "embedding_decoder", tgt_vocab_file, tgt_embed_file,
                    tgt_vocab_size, tgt_embed_size, dtype, embed_type=embed_type)

    return embedding_encoder, embedding_decoder


def _single_cell(unit_type, num_units, forget_bias, dropout, mode,
                 residual_connection=False, device_str=None, residual_fn=None):
    """Create an instance of a single RNN cell."""
    # dropout (= 1 - keep_prob) is set to 0 during eval and infer
    dropout = dropout if mode == tf.contrib.learn.ModeKeys.TRAIN else 0.0

    # Cell Type
    if unit_type == "lstm":
        utils.print_out("  LSTM, forget_bias=%g" % forget_bias, new_line=False)
        single_cell = tf.contrib.rnn.BasicLSTMCell(
            num_units,
            forget_bias=forget_bias)
    elif unit_type == "gru":
        utils.print_out("  GRU", new_line=False)
        single_cell = tf.contrib.rnn.GRUCell(num_units)
    elif unit_type == "layer_norm_lstm":
        utils.print_out("  Layer Normalized LSTM, forget_bias=%g" % forget_bias,
                        new_line=False)
        single_cell = tf.contrib.rnn.LayerNormBasicLSTMCell(
            num_units,
            forget_bias=forget_bias,
            layer_norm=True)
    elif unit_type == "nas":
        utils.print_out("  NASCell", new_line=False)
        single_cell = tf.contrib.rnn.NASCell(num_units)
    elif unit_type == "mlstm":
        utils.print_out("  Masked_LSTM, forget_bias=%g" % forget_bias, new_line=False)
        single_cell = tf.contrib.model_pruning.MaskedBasicLSTMCell(
            num_units,
            forget_bias=forget_bias)
    else:
        raise ValueError("Unknown unit type %s!" % unit_type)

    # Dropout (= 1 - keep_prob)
    if dropout > 0.0:
        single_cell = tf.contrib.rnn.DropoutWrapper(
            cell=single_cell, input_keep_prob=(1.0 - dropout))
        utils.print_out("  %s, dropout=%g " % (type(single_cell).__name__, dropout),
                        new_line=False)

    # Residual
    if residual_connection:
        single_cell = tf.contrib.rnn.ResidualWrapper(
            single_cell, residual_fn=residual_fn)
        utils.print_out("  %s" % type(single_cell).__name__, new_line=False)

    # Device Wrapper
    if device_str:
        single_cell = tf.contrib.rnn.DeviceWrapper(single_cell, device_str)
        utils.print_out("  %s, device=%s" %
                        (type(single_cell).__name__, device_str), new_line=False)

    return single_cell


def _cell_list(unit_type, num_units, num_layers, num_residual_layers,
               forget_bias, dropout, mode, num_gpus, base_gpu=0,
               single_cell_fn=None, residual_fn=None):
    """Create a list of RNN cells."""
    if not single_cell_fn:
        single_cell_fn = _single_cell

    # Multi-GPU
    cell_list = []
    for i in range(num_layers):
        utils.print_out("  cell %d" % i, new_line=False)
        single_cell = single_cell_fn(
            unit_type=unit_type,
            num_units=num_units,
            forget_bias=forget_bias,
            dropout=dropout,
            mode=mode,
            residual_connection=(i >= num_layers - num_residual_layers),
            device_str=get_device_str(i + base_gpu, num_gpus),
            residual_fn=residual_fn
        )
        utils.print_out("")
        cell_list.append(single_cell)

    return cell_list


def create_rnn_cell(unit_type, num_units, num_layers, num_residual_layers,
                    forget_bias, dropout, mode, num_gpus, base_gpu=0,
                    single_cell_fn=None):
    """Create multi-layer RNN cell.

    Args:
      unit_type: string representing the unit type, i.e. "lstm".
      num_units: the depth of each unit.
      num_layers: number of cells.
      num_residual_layers: Number of residual layers from top to bottom. For
        example, if `num_layers=4` and `num_residual_layers=2`, the last 2 RNN
        cells in the returned list will be wrapped with `ResidualWrapper`.
      forget_bias: the initial forget bias of the RNNCell(s).
      dropout: floating point value between 0.0 and 1.0:
        the probability of dropout.  this is ignored if `mode != TRAIN`.
      mode: either tf.contrib.learn.TRAIN/EVAL/INFER
      num_gpus: The number of gpus to use when performing round-robin
        placement of layers.
      base_gpu: The gpu device id to use for the first RNN cell in the
        returned list. The i-th RNN cell will use `(base_gpu + i) % num_gpus`
        as its device id.
      single_cell_fn: allow for adding customized cell.
        When not specified, we default to model_helper._single_cell
    Returns:
      An `RNNCell` instance.
    """
    cell_list = _cell_list(unit_type=unit_type,
                           num_units=num_units,
                           num_layers=num_layers,
                           num_residual_layers=num_residual_layers,
                           forget_bias=forget_bias,
                           dropout=dropout,
                           mode=mode,
                           num_gpus=num_gpus,
                           base_gpu=base_gpu,
                           single_cell_fn=single_cell_fn)

    if len(cell_list) == 1:  # Single layer.
        return cell_list[0]
    else:  # Multi layers
        return tf.contrib.rnn.MultiRNNCell(cell_list)


def gradient_clip(gradients, max_gradient_norm):
    """Clipping gradients of a model."""
    clipped_gradients, gradient_norm = tf.clip_by_global_norm(
        gradients, max_gradient_norm)
    gradient_norm_summary = [tf.summary.scalar("grad_norm", gradient_norm)]
    gradient_norm_summary.append(
        tf.summary.scalar("clipped_gradient", tf.global_norm(clipped_gradients)))

    return clipped_gradients, gradient_norm_summary, gradient_norm


def print_variables_in_ckpt(ckpt_path):
    """Print a list of variables in a checkpoint together with their shapes."""
    utils.print_out("# Variables in ckpt %s" % ckpt_path)
    reader = tf.train.NewCheckpointReader(ckpt_path)
    variable_map = reader.get_variable_to_shape_map()
    for key in sorted(variable_map.keys()):
        utils.print_out("  %s: %s" % (key, variable_map[key]))


def load_model(model, ckpt_path, session, name):
    """Load model from a checkpoint."""
    start_time = time.time()
    try:
        model.saver.restore(session, ckpt_path)
    except tf.errors.NotFoundError as e:
        utils.print_out("Can't load checkpoint")
        print_variables_in_ckpt(ckpt_path)
        utils.print_out("%s" % str(e))
    session.run(tf.tables_initializer())
    utils.print_out(
        "  loaded %s model parameters from %s, time %.2fs" %
        (name, ckpt_path, time.time() - start_time))
    return model


def load_quantized_model(model, ckpt_path, session, name):
    """Loads quantized model and dequantizes variables"""
    start_time = time.time()
    dequant_ops = []
    for tsr in tf.trainable_variables():
        with tf.variable_scope(tsr.name.split(':')[0], reuse=True):
            quant_tsr = tf.get_variable('quantized', dtype=tf.qint8)
            min_range = tf.get_variable('min_range')
            max_range = tf.get_variable('max_range')
            dequant_ops.append(
                tsr.assign(tf.dequantize(quant_tsr, min_range, max_range, 'SCALED')))
    restore_list = [tsr for tsr in tf.global_variables() if tsr not in tf.trainable_variables()]

    saver = tf.train.Saver(restore_list)
    try:
        saver.restore(session, ckpt_path)
    except tf.errors.NotFoundError as e:
        utils.print_out("Can't load checkpoint")
        print_variables_in_ckpt(ckpt_path)
        utils.print_out("%s" % str(e))
    session.run(tf.tables_initializer())
    session.run(dequant_ops)
    utils.print_out(
        "  loaded %s model parameters from %s, time %.2fs" %
        (name, ckpt_path, time.time() - start_time))
    return model


def add_quatization_variables(model):
    """Add to graph quantization variables"""
    with model.graph.as_default():
        for tsr in tf.trainable_variables():
            with tf.variable_scope(tsr.name.split(':')[0]):
                output, min_range, max_range = tf.quantize(
                    tsr, tf.reduce_min(tsr), tf.reduce_max(tsr), tf.qint8, mode='SCALED')
                tf.get_variable('quantized', initializer=output, trainable=False,
                                collections=[_QUANTIZATION_COLLECTION])
                tf.get_variable('min_range', initializer=min_range, trainable=False,
                                collections=[_QUANTIZATION_COLLECTION])
                tf.get_variable('max_range', initializer=max_range, trainable=False,
                                collections=[_QUANTIZATION_COLLECTION])


def quantize_checkpoint(session, ckpt_path):
    """Quantize current loaded model and saves checkpoint in ckpt_path"""
    save_list = [tsr for tsr in tf.global_variables() if tsr not in tf.trainable_variables()]
    saver = tf.train.Saver(save_list)
    session.run(tf.variables_initializer(tf.get_collection(_QUANTIZATION_COLLECTION)))
    saver.save(session, ckpt_path)
    utils.print_out('Saved quantized checkpoint as %s' % ckpt_path)


def avg_checkpoints(model_dir, num_last_checkpoints, global_step,
                    global_step_name):
    """Average the last N checkpoints in the model_dir."""
    checkpoint_state = tf.train.get_checkpoint_state(model_dir)
    if not checkpoint_state:
        utils.print_out("# No checkpoint file found in directory: %s" % model_dir)
        return None

    # Checkpoints are ordered from oldest to newest.
    checkpoints = (
        checkpoint_state.all_model_checkpoint_paths[-num_last_checkpoints:])

    if len(checkpoints) < num_last_checkpoints:
        utils.print_out(
            "# Skipping averaging checkpoints because not enough checkpoints is "
            "avaliable."
        )
        return None

    avg_model_dir = os.path.join(model_dir, "avg_checkpoints")
    if not tf.gfile.Exists(avg_model_dir):
        utils.print_out(
            "# Creating new directory %s for saving averaged checkpoints." %
            avg_model_dir)
        tf.gfile.MakeDirs(avg_model_dir)

    utils.print_out("# Reading and averaging variables in checkpoints:")
    var_list = tf.contrib.framework.list_variables(checkpoints[0])
    var_values, var_dtypes = {}, {}
    for (name, shape) in var_list:
        if name != global_step_name:
            var_values[name] = np.zeros(shape)

    for checkpoint in checkpoints:
        utils.print_out("    %s" % checkpoint)
        reader = tf.contrib.framework.load_checkpoint(checkpoint)
        for name in var_values:
            tensor = reader.get_tensor(name)
            var_dtypes[name] = tensor.dtype
            var_values[name] += tensor

    for name in var_values:
        var_values[name] /= len(checkpoints)

    # Build a graph with same variables in the checkpoints, and save the averaged
    # variables into the avg_model_dir.
    with tf.Graph().as_default():
        tf_vars = [
            tf.get_variable(v, shape=var_values[v].shape, dtype=var_dtypes[name])
            for v in var_values]

        placeholders = [tf.placeholder(v.dtype, shape=v.shape) for v in tf_vars]
        assign_ops = [tf.assign(v, p) for (v, p) in zip(tf_vars, placeholders)]
        tf.Variable(global_step, name=global_step_name, trainable=False)
        saver = tf.train.Saver(tf.all_variables())

        with tf.Session() as sess:
            sess.run(tf.initialize_all_variables())
            for p, assign_op, (name, value) in zip(placeholders, assign_ops,
                                                   six.iteritems(var_values)):
                sess.run(assign_op, {p: value})

            # Use the built saver to save the averaged checkpoint. Only keep 1
            # checkpoint and the best checkpoint will be moved to avg_best_metric_dir.
            saver.save(
                sess,
                os.path.join(avg_model_dir, "translate.ckpt"))

    return avg_model_dir


def create_or_load_model(model, model_dir, session, name):
    """Create translation model and initialize or load parameters in session."""
    latest_ckpt = tf.train.latest_checkpoint(model_dir)
    if latest_ckpt:
        model = load_model(model, latest_ckpt, session, name)
    else:
        start_time = time.time()
        session.run(tf.global_variables_initializer())
        session.run(tf.tables_initializer())
        utils.print_out("  created %s model with fresh parameters, time %.2fs" %
                        (name, time.time() - start_time))

    global_step = model.global_step.eval(session=session)
    return model, global_step


def compute_perplexity(model, sess, name):
    """Compute perplexity of the output of the model.

    Args:
      model: model for compute perplexity.
      sess: tensorflow session to use.
      name: name of the batch.

    Returns:
      The perplexity of the eval outputs.
    """
    total_loss = 0
    total_predict_count = 0
    start_time = time.time()

    while True:
        try:
            output_tuple = model.eval(sess)
            total_loss += output_tuple.eval_loss * output_tuple.batch_size
            total_predict_count += output_tuple.predict_count
        except tf.errors.OutOfRangeError:
            break

    perplexity = utils.safe_exp(total_loss / total_predict_count)
    utils.print_time("  eval %s: perplexity %.2f" % (name, perplexity),
                     start_time)
    return perplexity



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import difflib
import logging
from typing import Set, List

from num2words import num2words

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


class ComputedRelationExtraction(RelationExtraction):
    """
    Extract Relation between two mentions according to computation and rule based algorithms
    """
    def extract_all_relations(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> Set[RelationType]:
        """
        Try to find if mentions has anyone or more of the relations this class support

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            Set[RelationType]: One or more of: RelationType.EXACT_STRING, RelationType.FUZZY_FIT,
                RelationType.FUZZY_HEAD_FIT, RelationType.SAME_HEAD_LEMMA,
                RelationType.SAME_HEAD_LEMMA_RELAX
        """
        relations = set()
        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str

        if StringUtils.is_pronoun(mention_x_str.lower()) or StringUtils.is_pronoun(
                mention_y_str.lower()):
            relations.add(RelationType.NO_RELATION_FOUND)
            return relations

        exact_rel = self.extract_exact_string(mention_x, mention_y)
        fuzzy_rel = self.extract_fuzzy_fit(mention_x, mention_y)
        fuzzy_head_rel = self.extract_fuzzy_head_fit(mention_x, mention_y)
        same_head_rel = self.extract_same_head_lemma(mention_x, mention_y)
        if exact_rel != RelationType.NO_RELATION_FOUND:
            relations.add(exact_rel)
        if fuzzy_rel != RelationType.NO_RELATION_FOUND:
            relations.add(fuzzy_rel)
        if fuzzy_head_rel != RelationType.NO_RELATION_FOUND:
            relations.add(fuzzy_head_rel)
        if same_head_rel != RelationType.NO_RELATION_FOUND:
            relations.add(same_head_rel)

        if len(relations) == 0:
            relations.add(RelationType.NO_RELATION_FOUND)

        return relations

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str

        if StringUtils.is_pronoun(mention_x_str.lower()) or StringUtils.is_pronoun(
                mention_y_str.lower()):
            return RelationType.NO_RELATION_FOUND

        if relation == RelationType.EXACT_STRING:
            return self.extract_exact_string(mention_x, mention_y)
        if relation == RelationType.FUZZY_FIT:
            return self.extract_fuzzy_fit(mention_x, mention_y)
        if relation == RelationType.FUZZY_HEAD_FIT:
            return self.extract_fuzzy_head_fit(mention_x, mention_y)
        if relation == RelationType.SAME_HEAD_LEMMA:
            is_same_lemma = self.extract_same_head_lemma(mention_x, mention_y)
            if is_same_lemma != RelationType.NO_RELATION_FOUND:
                return relation

        return RelationType.NO_RELATION_FOUND

    @staticmethod
    def extract_same_head_lemma(mention_x: MentionDataLight,
                                mention_y: MentionDataLight) -> RelationType:
        """
        Check if input mentions has same head lemma relation

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            RelationType.SAME_HEAD_LEMMA or RelationType.NO_RELATION_FOUND
        """
        if StringUtils.is_preposition(mention_x.mention_head_lemma) or \
                StringUtils.is_preposition(mention_y.mention_head_lemma) or \
                StringUtils.is_determiner(mention_x.mention_head_lemma) or \
                StringUtils.is_determiner(mention_y.mention_head_lemma):
            return RelationType.NO_RELATION_FOUND
        if mention_x.mention_head_lemma == mention_y.mention_head_lemma:
            return RelationType.SAME_HEAD_LEMMA
        return RelationType.NO_RELATION_FOUND

    @staticmethod
    def extract_fuzzy_head_fit(mention_x: MentionDataLight,
                               mention_y: MentionDataLight) -> RelationType:
        """
        Check if input mentions has fuzzy head fit relation

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            RelationType.FUZZY_HEAD_FIT or RelationType.NO_RELATION_FOUND
        """
        if StringUtils.is_preposition(mention_x.mention_head_lemma.lower()) or \
                StringUtils.is_preposition(mention_y.mention_head_lemma.lower()):
            return RelationType.NO_RELATION_FOUND

        mention_y_tokens = mention_y.tokens_str.split()
        mention_x_tokens = mention_x.tokens_str.split()
        if mention_x.mention_head in mention_y_tokens or \
                mention_y.mention_head in mention_x_tokens:
            return RelationType.FUZZY_HEAD_FIT
        return RelationType.NO_RELATION_FOUND

    @staticmethod
    def extract_fuzzy_fit(mention_x: MentionDataLight,
                          mention_y: MentionDataLight) -> RelationType:
        """
        Check if input mentions has fuzzy fit relation

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            RelationType.FUZZY_FIT or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        mention1_str = mention_x.tokens_str
        mention2_str = mention_y.tokens_str
        if difflib.SequenceMatcher(None, mention1_str, mention2_str).ratio() * 100 >= 90:
            relation = RelationType.FUZZY_FIT
            return relation

        # Convert numbers to words
        x_words = [num2words(int(w)).replace('-', ' ') if w.isdigit() else w for w in
                   mention1_str.split()]
        y_words = [num2words(int(w)).replace('-', ' ') if w.isdigit() else w for w in
                   mention2_str.split()]

        fuzzy_result = difflib.SequenceMatcher(None, ' '.join(x_words),
                                               ' '.join(y_words)).ratio() * 100 >= 85
        if fuzzy_result:
            relation = RelationType.FUZZY_FIT
        return relation

    @staticmethod
    def extract_exact_string(mention_x: MentionDataLight,
                             mention_y: MentionDataLight) -> RelationType:
        """
        Check if input mentions has exact string relation

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            RelationType.EXACT_STRING or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        mention1_str = mention_x.tokens_str
        mention2_str = mention_y.tokens_str
        if StringUtils.is_preposition(mention1_str.lower()) or \
                StringUtils.is_preposition(mention2_str.lower()):
            return relation

        if mention1_str.lower() == mention2_str.lower():
            relation = RelationType.EXACT_STRING

        return relation

    @staticmethod
    def get_supported_relations() -> List[RelationType]:
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.EXACT_STRING, RelationType.FUZZY_FIT, RelationType.FUZZY_HEAD_FIT,
                RelationType.SAME_HEAD_LEMMA]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# pylint: disable=invalid-name
import numpy as np


# Things that were changed from the original:
# - Reformatted code and variable names to conform with PEP8
# - Added legal header


# This file contains routines from Lisbon Machine Learning summer school.
# The code is freely distributed under a MIT license. https://github.com/LxMLS/lxmls-toolkit/


def parse_proj(scores, gold=None):
    # pylint: disable=too-many-locals
    """
    Parse using Eisner's algorithm.
    """
    nr, nc = np.shape(scores)
    if nr != nc:
        raise ValueError("scores must be a squared matrix with nw+1 rows")

    N = nr - 1  # Number of words (excluding root).

    # Initialize CKY table.
    complete = np.zeros([N + 1, N + 1, 2])  # s, t, direction (right=1).
    incomplete = np.zeros([N + 1, N + 1, 2])  # s, t, direction (right=1).
    complete_backtrack = -np.ones([N + 1, N + 1, 2], dtype=int)  # s, t, direction (right=1).
    incomplete_backtrack = -np.ones([N + 1, N + 1, 2], dtype=int)  # s, t, direction (right=1).

    incomplete[0, :, 0] -= np.inf

    # Loop from smaller items to larger items.
    for k in range(1, N + 1):
        for s in range(N - k + 1):
            t = s + k

            # First, create incomplete items.
            # left tree
            incomplete_vals0 = complete[s, s:t, 1] + complete[(s + 1):(t + 1), t, 0] + scores[
                t, s] + (0.0 if gold is not None and gold[s] == t else 1.0)
            incomplete[s, t, 0] = np.max(incomplete_vals0)
            incomplete_backtrack[s, t, 0] = s + np.argmax(incomplete_vals0)
            # right tree
            incomplete_vals1 = complete[s, s:t, 1] + complete[(s + 1):(t + 1), t, 0] + scores[
                s, t] + (0.0 if gold is not None and gold[t] == s else 1.0)
            incomplete[s, t, 1] = np.max(incomplete_vals1)
            incomplete_backtrack[s, t, 1] = s + np.argmax(incomplete_vals1)

            # Second, create complete items.
            # left tree
            complete_vals0 = complete[s, s:t, 0] + incomplete[s:t, t, 0]
            complete[s, t, 0] = np.max(complete_vals0)
            complete_backtrack[s, t, 0] = s + np.argmax(complete_vals0)
            # right tree
            complete_vals1 = incomplete[s, (s + 1):(t + 1), 1] + complete[(s + 1):(t + 1), t, 1]
            complete[s, t, 1] = np.max(complete_vals1)
            complete_backtrack[s, t, 1] = s + 1 + np.argmax(complete_vals1)

    # value = complete[0][N][1]
    heads = [-1 for _ in range(N + 1)]  # -np.ones(N+1, dtype=int)
    _backtrack_eisner(incomplete_backtrack, complete_backtrack, 0, N, 1, 1, heads)
    value_proj = 0.0
    for m in range(1, N + 1):
        h = heads[m]
        value_proj += scores[h, m]
    return heads


# pylint: disable=too-many-arguments
def _backtrack_eisner(incomplete_backtrack, complete_backtrack, s, t, direction, complete, heads):
    """
    Backtracking step in Eisner's algorithm.
    - incomplete_backtrack is a (NW+1)-by-(NW+1) numpy array indexed by a start
     position, an end position, and a direction flag (0 means left, 1 means
     right). This array contains the arg-maxes of each step in the Eisner
     algorithm when building *incomplete* spans.
    - complete_backtrack is a (NW+1)-by-(NW+1) numpy array indexed by a start
     position, an end position, and a direction flag (0 means left, 1 means
     right). This array contains the arg-maxes of each step in the Eisner
     algorithm when building *complete* spans.
    - s is the current start of the span
    - t is the current end of the span
    - direction is 0 (left attachment) or 1 (right attachment)
    - complete is 1 if the current span is complete, and 0 otherwise
    - heads is a (NW+1)-sized numpy array of integers which is a placeholder
      for storing the head of each word.
    """
    if s == t:
        return
    if complete:
        r = complete_backtrack[s][t][direction]
        if direction == 0:
            _backtrack_eisner(incomplete_backtrack, complete_backtrack, s, r, 0, 1, heads)
            _backtrack_eisner(incomplete_backtrack, complete_backtrack, r, t, 0, 0, heads)
            return
        _backtrack_eisner(incomplete_backtrack, complete_backtrack, s, r, 1, 0, heads)
        _backtrack_eisner(incomplete_backtrack, complete_backtrack, r, t, 1, 1, heads)
        return

    r = incomplete_backtrack[s][t][direction]
    if direction == 0:
        heads[s] = t
        _backtrack_eisner(incomplete_backtrack, complete_backtrack, s, r, 1, 1, heads)
        _backtrack_eisner(incomplete_backtrack, complete_backtrack, r + 1, t, 0, 1, heads)
        return
    heads[t] = s
    _backtrack_eisner(incomplete_backtrack, complete_backtrack, s, r, 1, 1, heads)
    _backtrack_eisner(incomplete_backtrack, complete_backtrack, r + 1, t, 0, 1, heads)
    return



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np
import pickle
from os import makedirs, path, sys

from nlp_architect.api.abstract_api import AbstractApi
from nlp_architect.models.intent_extraction import MultiTaskIntentModel, Seq2SeqIntentModel
from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.generic import pad_sentences
from nlp_architect.utils.io import download_unlicensed_file
from nlp_architect.utils.text import SpacyInstance, bio_to_spans


class IntentExtractionApi(AbstractApi):
    model_dir = str(LIBRARY_OUT / 'intent-pretrained')
    pretrained_model_info = path.join(model_dir, 'model_info.dat')
    pretrained_model = path.join(model_dir, 'model.h5')

    def __init__(self, prompt=True):
        self.model = None
        self.model_type = None
        self.word_vocab = None
        self.tags_vocab = None
        self.char_vocab = None
        self.intent_vocab = None
        self._download_pretrained_model(prompt)
        self.nlp = SpacyInstance(disable=['tagger', 'ner', 'parser', 'vectors', 'textcat'])

    def process_text(self, text):
        input_text = ' '.join(text.strip().split())
        return self.nlp.tokenize(input_text)

    @staticmethod
    def _prompt():
        response = input('\nTo download \'{}\', please enter YES: '.
                         format('intent_extraction'))
        res = response.lower().strip()
        if res == "yes" or (len(res) == 1 and res == 'y'):
            print('Downloading {}...'.format('ner'))
            responded_yes = True
        else:
            print('Download declined. Response received {} != YES|Y. '.format(res))
            responded_yes = False
        return responded_yes

    @staticmethod
    def _download_pretrained_model(prompt=True):
        """Downloads the pre-trained BIST model if non-existent."""
        model_info_exists = path.isfile(IntentExtractionApi.pretrained_model_info)
        model_exists = path.isfile(IntentExtractionApi.pretrained_model)
        if not model_exists or not model_info_exists:
            print('The pre-trained models to be downloaded for the intent extraction dataset '
                  'are licensed under Apache 2.0. By downloading, you accept the terms '
                  'and conditions provided by the license')
            makedirs(IntentExtractionApi.model_dir, exist_ok=True)
            if prompt is True:
                agreed = IntentExtractionApi._prompt()
                if agreed is False:
                    sys.exit(0)
            download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data'
                                     '/models/intent/',
                                     'model_info.dat', IntentExtractionApi.pretrained_model_info)
            download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data'
                                     '/models/intent/',
                                     'model.h5', IntentExtractionApi.pretrained_model)
            print('Done.')

    @staticmethod
    def display_results(text_str, predictions, intent_type):
        ret = {'annotation_set': [], 'doc_text': ' '.join([t for t in text_str])}
        spans = []
        available_tags = set()
        for s, e, tag in bio_to_spans(text_str, predictions):
            spans.append({
                'start': s,
                'end': e,
                'type': tag
            })
            available_tags.add(tag)
        ret['annotation_set'] = list(available_tags)
        ret['spans'] = spans
        ret['title'] = intent_type
        return {'doc': ret, 'type': 'high_level'}

    def vectorize(self, doc, vocab, char_vocab=None):
        words = np.asarray([vocab[w.lower()] if w.lower() in vocab else 1 for w in doc])\
            .reshape(1, -1)
        if char_vocab is not None:
            sentence_chars = []
            for w in doc:
                word_chars = []
                for c in w:
                    if c in char_vocab:
                        _cid = char_vocab[c]
                    else:
                        _cid = 1
                    word_chars.append(_cid)
                sentence_chars.append(word_chars)
            sentence_chars = np.expand_dims(pad_sentences(sentence_chars, self.model.word_length),
                                            axis=0)
            return [words, sentence_chars]
        return words

    def inference(self, doc):
        text_arr = self.process_text(doc)
        intent_type = None
        if self.model_type == 'mtl':
            doc_vec = self.vectorize(text_arr, self.word_vocab, self.char_vocab)
            intent, tags = self.model.predict(doc_vec, batch_size=1)
            intent = int(intent.argmax(1).flatten())
            intent_type = self.intent_vocab.get(intent, None)
            print('Detected intent type: {}'.format(intent_type))
        else:
            doc_vec = self.vectorize(text_arr, self.word_vocab, None)
            tags = self.model.predict(doc_vec, batch_size=1)
        tags = tags.argmax(2).flatten()
        tag_str = [self.tags_vocab.get(n, None) for n in tags]
        for t, n in zip(text_arr, tag_str):
            print('{}\t{}\t'.format(t, n))
        return self.display_results(text_arr, tag_str, intent_type)

    def load_model(self):
        with open(IntentExtractionApi.pretrained_model_info, 'rb') as fp:
            model_info = pickle.load(fp)
        self.model_type = model_info['type']
        self.word_vocab = model_info['word_vocab']
        self.tags_vocab = {v: k for k, v in model_info['tags_vocab'].items()}
        if self.model_type == 'mtl':
            self.char_vocab = model_info['char_vocab']
            self.intent_vocab = {v: k for k, v in model_info['intent_vocab'].items()}
            model = MultiTaskIntentModel()
        else:
            model = Seq2SeqIntentModel()
        model.load(self.pretrained_model)
        self.model = model



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import pickle
from os import makedirs, path, sys

import numpy as np

from nlp_architect.api.abstract_api import AbstractApi
from nlp_architect.models.ner_crf import NERCRF
from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.generic import pad_sentences
from nlp_architect.utils.io import download_unlicensed_file
from nlp_architect.utils.text import SpacyInstance, bio_to_spans


class NerApi(AbstractApi):
    """
    NER model API
    """
    model_dir = str(LIBRARY_OUT / 'ner-pretrained')
    pretrained_model = path.join(model_dir, 'model_v4.h5')
    pretrained_model_info = path.join(model_dir, 'model_info_v4.dat')

    def __init__(self, prompt=True):
        self.model = None
        self.model_info = None
        self.word_vocab = None
        self.y_vocab = None
        self.char_vocab = None
        self._download_pretrained_model(prompt)
        self.nlp = SpacyInstance(disable=['tagger', 'ner', 'parser', 'vectors', 'textcat'])

    @staticmethod
    def _prompt():
        response = input('\nTo download \'{}\', please enter YES: '.
                         format('ner'))
        res = response.lower().strip()
        if res == "yes" or (len(res) == 1 and res == 'y'):
            print('Downloading {}...'.format('ner'))
            responded_yes = True
        else:
            print('Download declined. Response received {} != YES|Y. '.format(res))
            responded_yes = False
        return responded_yes

    def _download_pretrained_model(self, prompt=True):
        """Downloads the pre-trained BIST model if non-existent."""
        model_exists = path.isfile(self.pretrained_model)
        model_info_exists = path.isfile(self.pretrained_model_info)
        if not model_exists or not model_info_exists:
            print('The pre-trained models to be downloaded for the NER dataset '
                  'are licensed under Apache 2.0. By downloading, you accept the terms '
                  'and conditions provided by the license')
            makedirs(self.model_dir, exist_ok=True)
            if prompt is True:
                agreed = NerApi._prompt()
                if agreed is False:
                    sys.exit(0)
            download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data'
                                     '/models/ner/',
                                     'model_v4.h5', self.pretrained_model)
            download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data'
                                     '/models/ner/',
                                     'model_info_v4.dat', self.pretrained_model_info)
            print('Done.')

    def load_model(self):
        self.model = NERCRF()
        self.model.load(self.pretrained_model)
        with open(self.pretrained_model_info, 'rb') as fp:
            model_info = pickle.load(fp)
        self.word_vocab = model_info['word_vocab']
        self.y_vocab = {v: k for k, v in model_info['y_vocab'].items()}
        self.char_vocab = model_info['char_vocab']

    @staticmethod
    def pretty_print(text, tags):
        spans = []
        for s, e, tag in bio_to_spans(text, tags):
            spans.append({
                'start': s,
                'end': e,
                'type': tag
            })
        ents = dict((obj['type'].lower(), obj) for obj in spans).keys()
        ret = {'doc_text': ' '.join(text),
               'annotation_set': list(ents),
               'spans': spans,
               'title': 'None'}
        print({"doc": ret, 'type': 'high_level'})
        return {"doc": ret, 'type': 'high_level'}

    def process_text(self, text):
        input_text = ' '.join(text.strip().split())
        return self.nlp.tokenize(input_text)

    def vectorize(self, doc, vocab, char_vocab):
        words = np.asarray([vocab[w.lower()] if w.lower() in vocab else 1 for w in doc]) \
            .reshape(1, -1)
        sentence_chars = []
        for w in doc:
            word_chars = []
            for c in w:
                if c in char_vocab:
                    _cid = char_vocab[c]
                else:
                    _cid = 1
                word_chars.append(_cid)
            sentence_chars.append(word_chars)
        sentence_chars = np.expand_dims(pad_sentences(sentence_chars, self.model.word_length),
                                        axis=0)
        return words, sentence_chars

    def inference(self, doc):
        text_arr = self.process_text(doc)
        doc_vec = self.vectorize(text_arr, self.word_vocab, self.char_vocab)
        seq_len = np.array([len(text_arr)]).reshape(-1, 1)
        inputs = list(doc_vec)
        # pylint: disable=no-member
        inputs = list(doc_vec) + [seq_len]
        doc_ner = self.model.predict(inputs, batch_size=1).argmax(2).flatten()
        tags = [self.y_vocab.get(n, None) for n in doc_ner]
        return self.pretty_print(text_arr, tags)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import tensorflow as tf

from nlp_architect.utils.metrics import get_conll_scores


class ConllCallback(tf.keras.callbacks.Callback):
    """
    A Tensorflow(Keras) Conlleval evaluator.
    Runs the conlleval script for given x and y inputs.
    Prints Conlleval F1 score on the end of each epoch.

    Args:
        x: features matrix
        y: labels matrix
        y_vocab (dict): int-to-str labels lexicon
        batch_size (:obj:`int`, optional): batch size
    """

    def __init__(self, x, y, y_vocab, batch_size=1):
        super(ConllCallback, self).__init__()
        self.x = x
        self.y = y
        self.y_vocab = {v: k for k, v in y_vocab.items()}
        self.bsz = batch_size

    def on_epoch_end(self, epoch, logs=None):
        predictions = self.model.predict(self.x, batch_size=self.bsz)
        stats = get_conll_scores(predictions, self.y, self.y_vocab)
        print()
        print('Conll eval: \n{}'.format(stats))



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from typing import List

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType


class RelationExtraction(object):
    def __init__(self):
        pass

    def extract_relation(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                         relation: RelationType) -> RelationType:
        """
        Base Class Check if Sub class support given relation before executing the sub class

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation and
                RelationType.NO_RELATION_FOUND otherwise
        """
        ret_relation = RelationType.NO_RELATION_FOUND
        if relation in self.get_supported_relations():
            ret_relation = self.extract_sub_relations(mention_x, mention_y, relation)
        return ret_relation

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        raise NotImplementedError

    @staticmethod
    def get_supported_relations() -> List[RelationType]:
        raise NotImplementedError



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import re
import sys
from os import path
from typing import List, Tuple

import spacy
from nltk import WordNetLemmatizer
from nltk.stem.snowball import EnglishStemmer
from spacy.cli.download import download as spacy_download
from spacy.lang.en import LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES
from spacy.lemmatizer import Lemmatizer

from nlp_architect.utils.generic import license_prompt


class Vocabulary:
    """
    A vocabulary that maps words to ints (storing a vocabulary)
    """

    def __init__(self, start=0):

        self._vocab = {}
        self._rev_vocab = {}
        self.next = start

    def add(self, word):
        """
        Add word to vocabulary

        Args:
            word (str): word to add

        Returns:
            int: id of added word
        """
        if word not in self._vocab.keys():
            self._vocab[word] = self.next
            self._rev_vocab[self.next] = word
            self.next += 1
        return self._vocab.get(word)

    def word_id(self, word):
        """
        Get the word_id of given word

        Args:
            word (str): word from vocabulary

        Returns:
            int: int id of word
        """
        return self._vocab.get(word, None)

    def __getitem__(self, item):
        """
        Get the word_id of given word (same as `word_id`)
        """
        return self.word_id(item)

    def __len__(self):
        return len(self._vocab)

    def __iter__(self):
        for word in self.vocab.keys():
            yield word

    @property
    def max(self):
        return self.next

    def id_to_word(self, wid):
        """
        Word-id to word (string)

        Args:
            wid (int): word id

        Returns:
            str: string of given word id
        """
        return self._rev_vocab.get(wid)

    @property
    def vocab(self):
        """
        dict: get the dict object of the vocabulary
        """
        return self._vocab

    def add_vocab_offset(self, offset):
        """
        Adds an offset to the ints of the vocabulary

        Args:
            offset (int): an int offset
        """
        new_vocab = {}
        for k, v in self.vocab.items():
            new_vocab[k] = v + offset
        self.next += offset
        self._vocab = new_vocab
        self._rev_vocab = {v: k for k, v in new_vocab.items()}

    def reverse_vocab(self):
        """
        Return the vocabulary as a reversed dict object

        Returns:
            dict: reversed vocabulary object
        """
        return self._rev_vocab


def try_to_load_spacy(model_name):
    try:
        spacy.load(model_name)
        return True
    except OSError:
        return False


class SpacyInstance:
    """
    Spacy pipeline wrapper which prompts user for model download authorization.

    Args:
        model (str, optional): spacy model name (default: english small model)
        disable (list of string, optional): pipeline annotators to disable
            (default: [])
        display_prompt (bool, optional): flag to display/skip license prompt
    """

    def __init__(self, model='en', disable=None, display_prompt=True):
        if disable is None:
            disable = []
        try:
            self._parser = spacy.load(model, disable=disable)
        except OSError:
            url = 'https://spacy.io/models'
            if display_prompt and license_prompt('Spacy {} model'.format(model), url) is False:
                sys.exit(0)
            spacy_download(model)
            self._parser = spacy.load(model, disable=disable)

    @property
    def parser(self):
        """return Spacy's instance parser"""
        return self._parser

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenize a sentence into tokens
        Args:
            text (str): text to tokenize

        Returns:
            list: a list of str tokens of input
        """
        # pylint: disable=not-callable

        return [t.text for t in self.parser(text)]


stemmer = EnglishStemmer()
lemmatizer = WordNetLemmatizer()
spacy_lemmatizer = Lemmatizer(LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES)
p = re.compile(r'[ \-,;.@&_]')


class Stopwords(object):
    """
    Stop words list class.
    """
    stop_words = []

    @staticmethod
    def get_words():
        if not Stopwords.stop_words:
            sw_path = path.join(path.dirname(path.realpath(__file__)),
                                'resources',
                                'stopwords.txt')
            with open(sw_path) as fp:
                stop_words = []
                for w in fp:
                    stop_words.append(w.strip().lower())
            Stopwords.stop_words = stop_words
        return Stopwords.stop_words


def simple_normalizer(text):
    """
    Simple text normalizer. Runs each token of a phrase thru wordnet lemmatizer
    and a stemmer.
    """
    if not str(text).isupper() or \
            not str(text).endswith('S') or \
            not len(text.split()) == 1:
        tokens = list(filter(lambda x: len(x) != 0, p.split(text.strip())))
        text = ' '.join([stemmer.stem(lemmatizer.lemmatize(t))
                         for t in tokens])
    return text


def spacy_normalizer(text, lemma=None):
    """
    Simple text normalizer using spacy lemmatizer. Runs each token of a phrase
    thru a lemmatizer and a stemmer.
    Arguments:
        text(string): the text to normalize.
        lemma(string): lemma of the given text. in this case only stemmer will
        run.
    """
    if not str(text).isupper() or \
            not str(text).endswith('S') or \
            not len(text.split()) == 1:
        tokens = list(filter(lambda x: len(x) != 0, p.split(text.strip())))
        if lemma:
            lemma = lemma.split(' ')
            text = ' '.join([stemmer.stem(l)
                             for l in lemma])
        else:
            text = ' '.join([stemmer.stem(spacy_lemmatizer(t, u'NOUN')[0])
                             for t in tokens])
    return text


def read_sequential_tagging_file(file_path, ignore_line_patterns=None):
    """
    Read a tab separated sequential tagging file.
    Returns a list of list of tuple of tags (sentences, words)

    Args:
        file_path (str): input file path
        ignore_line_patterns (list, optional): list of string patterns to ignore

    Returns:
        list of list of tuples
    """
    if ignore_line_patterns:
        assert isinstance(ignore_line_patterns, list), 'ignore_line_patterns must be a list'

    def _split_into_sentences(file_lines):
        sentences = []
        s = []
        for line in file_lines:
            if len(line) == 0:
                sentences.append(s)
                s = []
                continue
            s.append(line)
        if len(s) > 0:
            sentences.append(s)
        return sentences

    with open(file_path, encoding='utf-8') as fp:
        data = fp.readlines()
        data = [d.strip() for d in data]
        if ignore_line_patterns:
            for s in ignore_line_patterns:
                data = [d for d in data if s not in d]
        data = [tuple(d.split()) for d in data]
    return _split_into_sentences(data)


def word_vector_generator(data, lower=False, start=0):
    """
    Word vector generator util.
    Transforms a list of sentences into numpy int vectors and returns the
    constructed vocabulary

    Arguments:
        data (list): list of list of strings
        lower (bool, optional): transform strings into lower case
        start (int, optional): vocabulary index start integer

    Returns:
        2D numpy array and Vocabulary of the detected words
    """
    vocab = Vocabulary(start)
    data_vec = []
    for sentence in data:
        sentence_vec = []
        for w in sentence:
            word = w
            if lower:
                word = word.lower()
            wid = vocab[word]
            if wid is None:
                wid = vocab.add(word)
            sentence_vec.append(wid)
        data_vec.append(sentence_vec)
    return data_vec, vocab


def character_vector_generator(data, start=0):
    """
    Character word vector generator util.
    Transforms a list of sentences into numpy int vectors of the characters
    of the words of the sentence, and returns the constructed vocabulary

    Arguments:
        data (list): list of list of strings
        start (int, optional): vocabulary index start integer

    Returns:
        np.array: a 2D numpy array
        Vocabulary: constructed vocabulary
    """
    vocab = Vocabulary(start)
    data_vec = []
    for sentence in data:
        sentence_vec = []
        for w in sentence:
            word_vec = []
            for char in w:
                cid = vocab[char]
                if cid is None:
                    cid = vocab.add(char)
                word_vec.append(cid)
            sentence_vec.append(word_vec)
        data_vec.append(sentence_vec)
    return data_vec, vocab


def extract_nps(annotation_list, text=None):
    """
    Extract Noun Phrases from given text tokens and phrase annotations.
    Returns a list of tuples with start/end indexes.

    Args:
        annotation_list (list): a list of annotation tags in str
        text (list, optional): a list of token texts in str

    Returns:
        list of start/end markers of noun phrases, if text is provided a list of noun phrase texts
    """
    np_starts = [i for i in range(len(annotation_list)) if annotation_list[i] == 'B-NP']
    np_markers = []
    for s in np_starts:
        i = 1
        while s + i < len(annotation_list) and annotation_list[s + i] == 'I-NP':
            i += 1
        np_markers.append((s, s + i))
    return_markers = np_markers
    if text:
        assert len(text) == len(annotation_list), 'annotations/text length mismatch'
        return_markers = [' '.join(text[s:e]) for s, e in np_markers]
    return return_markers


def bio_to_spans(text: List[str], tags: List[str]) -> List[Tuple[int, int, str]]:
    """
    Convert BIO tagged list of strings into span starts and ends
    Args:
        text: list of words
        tags: list of tags

    Returns:
        tuple: list of start, end and tag of detected spans
    """
    pointer = 0
    starts = []
    for i, t, in enumerate(tags):
        if t.startswith('B-'):
            starts.append((i, pointer))
        pointer += len(text[i]) + 1

    spans = []
    for s_i, s_char in starts:
        label_str = tags[s_i][2:]
        e = 0
        e_char = len(text[s_i + e])
        while len(tags) > s_i + e + 1 and tags[s_i + e + 1].startswith('I-'):
            e += 1
            e_char += 1 + len(text[s_i + e])
        spans.append((s_char, s_char + e_char, label_str))
    return spans



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from typing import Set, Dict


class WordnetPage(object):
    def __init__(self, orig_phrase: str, clean_phrase: str, head: str, head_lemma: str,
                 head_synonyms: Set[str], head_lemma_synonyms: Set[str],
                 head_derivationally: Set[str], head_lemma_derivationally: Set[str],
                 all_clean_words_synonyms: Set[str]) -> None:
        """
        Object represent a Wikipedia Page and extracted fields.

        Args:
            orig_phrase (str): original search phrase
            clean_phrase (str): original search phrase normalized
            head (str): page title head
            head_lemma (str): page title head lemma
            head_synonyms (set): head synonyms words extracted from wordnet
            head_lemma_synonyms (set): head lemma synonyms words extracted from wordnet
            head_derivationally (set): wordnet head derivationally_related_forms()
            head_lemma_derivationally (set): wordnet head lemma derivationally_related_forms()
            all_clean_words_synonyms (set): clean_phrase wordnet synonyms
        """
        self.orig_phrase = orig_phrase
        self.clean_phrase = clean_phrase
        self.head = head
        self.head_lemma = head_lemma
        self.head_synonyms = head_synonyms
        self.head_lemma_synonyms = head_lemma_synonyms
        self.head_derivationally = head_derivationally
        self.head_lemma_derivationally = head_lemma_derivationally
        self.all_clean_words_synonyms = all_clean_words_synonyms

    def __eq__(self, other):
        return self.orig_phrase == other.orig_phrase and self.head == other.head and \
            self.head_lemma == other.head_lemma

    def __hash__(self):
        return hash(self.orig_phrase) + hash(self.head) + hash(self.head_lemma)

    def toJson(self) -> Dict:
        result_dict = dict()
        result_dict['orig_phrase'] = self.orig_phrase
        result_dict['clean_phrase'] = self.clean_phrase
        result_dict['head'] = self.head
        result_dict['head_lemma'] = self.head_lemma
        result_dict['head_synonyms'] = list(self.head_synonyms)
        result_dict['head_lemma_synonyms'] = list(self.head_lemma_synonyms)
        result_dict['head_derivationally'] = list(self.head_derivationally)
        result_dict['head_lemma_derivationally'] = list(self.head_lemma_derivationally)
        if self.all_clean_words_synonyms is not None:
            all_as_list = []
            for set_ in self.all_clean_words_synonyms:
                all_as_list.append(list(set_))
            result_dict['all_clean_words_synonyms'] = all_as_list

        return result_dict



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
from typing import List

from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.models.cross_doc_coref.sieves_config import \
    EventSievesConfiguration, EntitySievesConfiguration

logger = logging.getLogger(__name__)


class SievesContainerInitialization(object):
    def __init__(self, event_coref_config: EventSievesConfiguration,
                 entity_coref_config: EntitySievesConfiguration,
                 sieves_model_list: List[RelationExtraction]):
        self.sieves_model_list = sieves_model_list
        self.event_config = event_coref_config
        self.entity_config = entity_coref_config

    def get_module_from_relation(self, relation_type):
        for model in self.sieves_model_list:
            if relation_type in model.get_supported_relations():
                return model

        raise Exception('No model found that Support RelationType-' + str(relation_type))



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import logging
from typing import List

from nlp_architect.common.cdc.cluster import Clusters
from nlp_architect.common.cdc.topics import Topics
from nlp_architect.models.cross_doc_coref.system.sieves.run_sieve_system import get_run_system
from nlp_architect.models.cross_doc_coref.system.sieves_container_init import \
    SievesContainerInitialization

logger = logging.getLogger(__name__)


def run_event_coref(topics: Topics, resources: SievesContainerInitialization) -> List[Clusters]:
    """
    Running Cross Document Coref on event mentions
    Args:
        topics   : The Topics (with mentions) to evaluate
        resources: resources for running the evaluation

    Returns:
        Clusters: List of clusters and mentions with predicted cross doc coref within each topic
    """

    return _run_coref(topics, resources, 'event')


def run_entity_coref(topics: Topics, resources: SievesContainerInitialization) -> List[Clusters]:
    """
    Running Cross Document Coref on Entity mentions
    Args:
        topics   : The Topics (with mentions) to evaluate
        resources: (SievesContainerInitialization) resources for running the evaluation

    Returns:
        Clusters: List of topics and mentions with predicted cross doc coref within each topic
    """
    return _run_coref(topics, resources, 'entity')


def _run_coref(topics: Topics, resources: SievesContainerInitialization,
               eval_type: str) -> List[Clusters]:
    """
    Running Cross Document Coref on Entity mentions
    Args:
        resources: (SievesContainerInitialization) resources for running the evaluation
        topics   : The Topics (with mentions) to evaluate

    Returns:
        Clusters: List of topics and mentions with predicted cross doc coref within each topic
    """
    clusters_list = list()
    for topic in topics.topics_list:
        sieves_list = get_run_system(topic, resources, eval_type)
        clusters = sieves_list.run_deterministic()
        clusters.set_coref_chain_to_mentions()
        clusters_list.append(clusters)

    return clusters_list



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np
import pandas as pd

from bokeh.core.properties import value
from bokeh.document import Document
from bokeh.transform import dodge
from bokeh.layouts import widgetbox, row
from bokeh import layouts
from bokeh.plotting import figure, Figure
from bokeh.models.widgets import DataTable, TableColumn, HTMLTemplateFormatter, RadioButtonGroup
from bokeh.models import HoverTool, Title, Div
from bokeh.layouts import column
from bokeh.models import ColumnDataSource
from bokeh.server.server import Server

from nlp_architect.models.absa.inference.data_types import SentimentDoc, SentimentSentence

POLARITIES = ('POS', 'NEG')


def serve_ui(stats: pd.DataFrame, aspects: pd.Series) -> None:
    """Main function for serving UI application.

    Args:
        stats (pd.DataFrame): Table containing aggregated stats for aspect-polarity pairs
        aspects (pd.Series): List of all aspects
    """
    def _doc_modifier(doc: Document) -> None:
        plot, source = _create_plot(stats, aspects)
        events_table = _create_events_table()
        events_type = RadioButtonGroup(labels=['All Events', 'In-Domain Events'], active=0)

        # pylint: disable=unused-argument
        def _events_handler(attr, old, new):
            _update_events(stats, aspects, events_table, source, events_type.active)

        events_type.on_change('active', _events_handler)
        source.selected.on_change('indices', _events_handler)  # pylint: disable=no-member
        doc.add_root(column(_create_header(), plot, events_type, events_table,
                            sizing_mode="scale_width"))

    print('Opening Bokeh application on http://localhost:5006/')
    server = Server({'/': _doc_modifier})
    server.start()
    server.io_loop.add_callback(server.show, "/")
    server.io_loop.start()


def _create_events_table() -> DataTable:
    """Utility function for creating and styling the events table."""
    formatter = HTMLTemplateFormatter(template='''
    <style>
        .AS_POS {color: #0000FF; font-weight: bold;}
        .AS_NEG {color: #0000FF; font-weight: bold;}
        .OP_POS {color: #1aaa0d; font-style: bold;}
        .OP_NEG {color: #f40000;font-style: bold;}
        .NEG_POS {font-style: italic;}
        .NEG_NEG {color: #f40000; font-style: italic;}
        .INT_POS {color: #1aaa0d; font-style: italic;}
        .INT_NEG {color: #f40000; font-style: italic;}
        * {font-size: 105%;}
    </style>
    <%= value %>''')
    columns = [TableColumn(field='POS_events', title='Positive Events', formatter=formatter),
               TableColumn(field='NEG_events', title='Negative Events', formatter=formatter)]
    return DataTable(source=ColumnDataSource(), columns=columns, height=400, index_position=None,
                     width=1500, sortable=False, editable=False, reorderable=False)


def _create_plot(stats: pd.DataFrame, aspects: pd.Series) -> (Figure, ColumnDataSource):
    """Utility function for creating and styling the bar plot."""
    pos_counts, neg_counts = \
        ([stats.loc[(asp, pol, False), 'Quantity'] for asp in aspects] for pol in POLARITIES)
    np.seterr(divide='ignore')
    source = ColumnDataSource(data={'aspects': aspects, 'POS': pos_counts, 'NEG': neg_counts,
                                    'log-POS': np.log2(pos_counts),
                                    'log-NEG': np.log2(neg_counts)})
    np.seterr(divide='warn')
    p = figure(plot_height=150, sizing_mode="scale_width",
               x_range=aspects, toolbar_location='left', tools='save,tap,pan,zoom_in,zoom_out')
    rs = [p.vbar(x=dodge('aspects', -0.207, range=p.x_range), top='log-POS', width=0.4,
                 source=source, color="limegreen", legend=value('POS'), name='POS'),
          p.vbar(x=dodge('aspects', 0.207, range=p.x_range), top='log-NEG', width=0.4,
                 source=source, color="orangered", legend=value('NEG'), name='NEG')]
    for r in rs:
        p.add_tools(HoverTool(tooltips=[('Aspect', '@aspects'), (r.name, '@' + r.name)],
                              renderers=[r]))
    p.add_layout(Title(text=' ' * 5 + 'Log Scale', align='left', text_font_size='20px'), 'left')
    p.yaxis.ticker = []
    p.y_range.start = 0
    p.xgrid.grid_line_color = None
    return p, source


def _update_events(stats: pd.DataFrame, aspects: pd.Series, events: DataTable,
                   source: ColumnDataSource, in_domain: bool) -> None:
    """Utility function for updating the content of the events table."""
    i = source.selected.indices
    events.source.data.update({pol + '_events': stats.loc[aspects[i[0]], pol, in_domain]
                               ['Sent_1':].replace(np.nan, '') if i else [] for pol in POLARITIES})


def _create_header() -> layouts.Row:
    """Utility function for creating and styling the header row in the UI layout."""
    lib_github = Div(text="<a href='https://github.com/NervanaSystems/nlp-architect' "
                     "style='text-decoration:none'>NLP ARCHITECT by IntelAI</a>",
                     style={'font-size': '140%', 'color': 'darkblue', 'font-weight': 'bold'})
    title = Div(text="Aspect-Based Sentiment Analysis",
                style={'font-size': '200%', 'color': 'royalblue', 'font-weight': 'bold'})
    return row(widgetbox(lib_github, width=600), widgetbox(title, width=700))


def _ui_format(sent: SentimentSentence, doc: SentimentDoc) -> str:
    """Get sentence as HTML with 4 classes: aspects, opinions, negations and intensifiers."""
    text = doc.doc_text[sent.start: sent.end + 1]
    seen = set()
    for term in sorted([t for e in sent.events for t in e], key=lambda t: t.start)[::-1]:
        if term.start not in seen:
            seen.add(term.start)
            start = term.start - sent.start
            end = start + term.len
            label = term.type.value + '_' + term.polarity.value
            text = ''.join((text[:start], '<span class="', label, '">', text[start: end],
                            '</span>', text[end:]))
    return text



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from typing import List, Tuple

from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType


class SievesConfiguration(object):
    def __init__(self):
        """Cross document co-reference event and entity evaluation configuration settings"""

        self.__sieves_order = None
        self.__run_evaluation = False

    @property
    def sieves_order(self):
        """
        Sieve definition and Sieve running order

        Tuple[SieveType, RelationType, Threshold(float)] - define sieves to run, were

        Strict- Merge clusters only in case all mentions has current relation between them,
        Relax- Merge clusters in case (matched mentions) / len(cluster_1.mentions)) >= thresh,
        Very_Relax- Merge clusters in case (matched mentions) / (all possible pairs) >= thresh

        RelationType represent the type of sieve to run.

        """
        return self.__sieves_order

    @sieves_order.setter
    def sieves_order(self, sieves_order: List[Tuple[RelationType, float]]):
        self.__sieves_order = sieves_order

    @property
    def run_evaluation(self):
        """Should run evaluation (True/False)"""
        return self.__run_evaluation

    @run_evaluation.setter
    def run_evaluation(self, run_evaluation: bool):
        self.__run_evaluation = run_evaluation


class EventSievesConfiguration(SievesConfiguration):
    def __init__(self):
        super(EventSievesConfiguration, self).__init__()

        self.run_evaluation = True

        self.sieves_order = [
            (RelationType.SAME_HEAD_LEMMA, 1.0),
            (RelationType.EXACT_STRING, 1.0),
            (RelationType.WIKIPEDIA_DISAMBIGUATION, 0.1),
            (RelationType.WORD_EMBEDDING_MATCH, 0.7),
            (RelationType.WIKIPEDIA_REDIRECT_LINK, 0.1),
            (RelationType.FUZZY_HEAD_FIT, 0.5),
            (RelationType.FUZZY_FIT, 1.0),
            (RelationType.WITHIN_DOC_COREF, 1.0),
            (RelationType.WIKIPEDIA_TITLE_PARENTHESIS, 0.1),
            (RelationType.WIKIPEDIA_BE_COMP, 0.1),
            (RelationType.WIKIPEDIA_CATEGORY, 0.1),
            (RelationType.VERBOCEAN_MATCH, 0.1),
            (RelationType.WORDNET_DERIVATIONALLY, 1.0)
        ]


class EntitySievesConfiguration(SievesConfiguration):
    def __init__(self):
        super(EntitySievesConfiguration, self).__init__()

        self.run_evaluation = True

        self.sieves_order = [
            (RelationType.SAME_HEAD_LEMMA, 1.0),
            (RelationType.EXACT_STRING, 1.0),
            (RelationType.FUZZY_FIT, 1.0),
            (RelationType.WIKIPEDIA_REDIRECT_LINK, 0.1),
            (RelationType.WIKIPEDIA_DISAMBIGUATION, 0.1),
            (RelationType.WORD_EMBEDDING_MATCH, 0.7),
            (RelationType.WORDNET_PARTIAL_SYNSET_MATCH, 0.1),
            (RelationType.FUZZY_HEAD_FIT, 0.5),
            (RelationType.WIKIPEDIA_CATEGORY, 0.1),
            (RelationType.WITHIN_DOC_COREF, 1.0),
            (RelationType.WIKIPEDIA_BE_COMP, 0.1),
            (RelationType.WIKIPEDIA_TITLE_PARENTHESIS, 0.1),
            (RelationType.WORDNET_SAME_SYNSET, 1.0),
            (RelationType.REFERENT_DICT, 0.5)
        ]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import os
import shutil
import subprocess
from http.server import HTTPServer as BaseHTTPServer, SimpleHTTPRequestHandler
from subprocess import run

import pytest

from nlp_architect import LIBRARY_ROOT, LIBRARY_PATH, LIBRARY_OUT
from nlp_architect.utils import ansi2html
from nlp_architect.version import NLP_ARCHITECT_VERSION


def run_cmd(command):
    return run(command.split(), shell=False)


class DocsCommand(object):
    cmd_name = 'doc'
    docs_source = LIBRARY_ROOT / 'doc'

    def __init__(self, subparsers):
        parser = subparsers.add_parser(DocsCommand.cmd_name,
                                       description='Build documentation',
                                       help='Build documentation')
        parser.set_defaults(func=DocsCommand.run_docs)
        self.parser = parser

    @staticmethod
    def run_docs(_):
        base_cmd = 'make -C {}'.format(DocsCommand.docs_source)
        print('Re-building documentation')
        run_cmd(base_cmd + ' clean')
        run_cmd(base_cmd + ' html')
        print('Documentation built in: {}'.format(DocsCommand.docs_source / 'build' / 'html'))
        print('To view documents point your browser to: http://localhost:8000')

        class HTTPHandler(SimpleHTTPRequestHandler):
            def translate_path(self, path):
                path = SimpleHTTPRequestHandler.translate_path(self, path)
                relpath = os.path.relpath(path, os.getcwd())
                fullpath = os.path.join(self.server.base_path, relpath)
                return fullpath

        class HTTPServer(BaseHTTPServer):
            def __init__(self, base_path, server_address, request_handler_class=HTTPHandler):
                self.base_path = base_path
                BaseHTTPServer.__init__(self, server_address, request_handler_class)

        web_dir = DocsCommand.docs_source / 'build' / 'html'
        httpd = HTTPServer(web_dir, ("", 8000))
        httpd.serve_forever()


class StyleCommand(object):
    cmd_name = 'style'
    check_dirs = [
        'examples',
        'nlp_architect',
        'tests'
    ]
    files_to_check = [str(LIBRARY_ROOT / f) for f in check_dirs]

    def __init__(self, subparsers):
        parser = subparsers.add_parser(StyleCommand.cmd_name,
                                       description='Run style checks on NLP Architect code',
                                       help='Run style checks on NLP Architect code')
        parser.add_argument('--only-flake', default=False, action='store_true',
                            help='Run only Flake8 style check')
        parser.add_argument('--only-pylint', default=False, action='store_true',
                            help='Run only Pylint style check')
        parser.set_defaults(func=StyleCommand.run_check)
        self.parser = parser

    @staticmethod
    def run_check(args):
        rc = 0
        run_all = not args.only_flake and not args.only_pylint
        if args.only_flake or run_all:
            f_err = StyleCommand.run_flake()
            if int(f_err) > 0:
                print('Flake8 errors found. RC={}'.format(f_err))
                rc = 1
        if args.only_pylint or run_all:
            p_err = StyleCommand.run_pylint()
            if int(p_err) != 0:
                print('pylint errors found. RC={}'.format(p_err))
                rc = 1
        if rc > 0:
            exit(1)

    @staticmethod
    def run_flake():
        print('Running flake8 ...\n')
        flake8_config = str(LIBRARY_ROOT / 'setup.cfg')
        os.makedirs(LIBRARY_OUT, exist_ok=True)
        flake8_out = str(LIBRARY_OUT / 'flake8.txt')
        flake8_html_out = str(LIBRARY_OUT / 'flake_html')
        print('HHH:' + flake8_html_out)
        try:
            os.remove(flake8_out)
            shutil.rmtree(flake8_html_out, ignore_errors=True)
        except OSError:
            pass
        cmd = 'flake8 {} --config={} --output-file {}' \
            .format(' '.join(StyleCommand.files_to_check), flake8_config, flake8_out)
        cmd_out = run_cmd(cmd)

        cmd_html = 'flake8 {} --config={} --format=html --htmldir={}' \
            .format(' '.join(StyleCommand.files_to_check), flake8_config, flake8_html_out)

        run_cmd(cmd_html)
        print('To view flake8 results in html, point your web browser to: \n{}\n'
              .format(flake8_html_out + '/index.html'))

        return cmd_out.returncode

    @staticmethod
    def run_pylint():
        print('Running pylint ...\n')
        pylint_config = LIBRARY_ROOT / 'pylintrc'
        os.makedirs(LIBRARY_OUT, exist_ok=True)
        pylint_out = str(LIBRARY_OUT / 'pylint.txt')
        html_out = str(LIBRARY_OUT / 'pylint.html')

        cmd = 'pylint -j 4 {} --rcfile {} --score=n'\
            .format(' '.join(StyleCommand.files_to_check), pylint_config)
        cmd_tee = 'tee {}'.format(pylint_out)
        ps = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        subprocess.run(cmd_tee.split(), stdin=ps.stdout)
        ret_code = 1 if os.stat(pylint_out).st_size else 0

        ansi2html.run(pylint_out, html_out)
        print('To view pylint results in html, point your web browser to: \n{}\n'.format(html_out))
        return ret_code


class TestCommand(object):
    cmd_name = 'test'

    def __init__(self, subparsers):
        parser = subparsers.add_parser(TestCommand.cmd_name,
                                       description='Run NLP Architect tests',
                                       help='Run NLP Architect tests')
        parser.add_argument('-f', type=str, help='test filename (runs test only of this file)')
        parser.set_defaults(func=TestCommand.run_tests)
        self.parser = parser

    @staticmethod
    def run_tests(args):
        # run all tests
        print('\nrunning NLP Architect tests ...')
        tests_dir = str(LIBRARY_ROOT / 'tests')
        tests = None
        if args.f:
            specific_test_file = args.f
            test_path = os.path.join(os.getcwd(), specific_test_file)
            if os.path.exists(test_path):
                tests = test_path
        else:
            tests = tests_dir
        if tests:
            TestCommand._prepare_tests()
            exit_code = pytest.main([tests,
                                     '-rs',
                                     '-vv',
                                     '--cov=nlp_architect',
                                     '--junit-xml=pytest_unit.xml'])
            print('tests exit code={}'.format(exit_code))
            if int(exit_code) != 0:
                exit(1)

    @staticmethod
    def _prepare_tests():
        import spacy
        from spacy.cli.download import download as spacy_download
        try:
            spacy.load('en')
        except OSError:
            spacy_download('en')

        from nlp_architect.api.machine_comprehension_api import MachineComprehensionApi
        from nlp_architect.api.intent_extraction_api import IntentExtractionApi
        from nlp_architect.api.ner_api import NerApi
        NerApi(prompt=False)
        IntentExtractionApi(prompt=False)
        MachineComprehensionApi(prompt=False).download_model()


class ServerCommand(object):
    cmd_name = 'server'

    def __init__(self, subparsers):
        parser = subparsers.add_parser(ServerCommand.cmd_name,
                                       description='Run NLP Architect server and demo UI',
                                       help='Run NLP Architect server and demo UI')
        parser.add_argument('-p', '--port', type=int, default=8080, help='server port')
        parser.set_defaults(func=ServerCommand.run_server)
        self.parser = parser

    @staticmethod
    def run_server(args):
        port = args.port
        serve_file = LIBRARY_PATH / 'server' / 'serve.py'
        cmd_str = 'hug -p {} -f {}'.format(port, serve_file)
        run_cmd(cmd_str)


# sub commands list
sub_commands = [
    TestCommand,
    StyleCommand,
    DocsCommand,
    ServerCommand,
]


def main():
    prog = 'nlp_architect'
    parser = argparse.ArgumentParser(description='NLP Architect runner', prog=prog)
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s v{}'.format(NLP_ARCHITECT_VERSION))

    subparsers = parser.add_subparsers(title='commands', metavar='')
    for command in sub_commands:
        command(subparsers)
    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# pylint: disable=deprecated-module
import os
import subprocess
from collections import Counter

from nlp_architect.data.conll import ConllEntry
from nlp_architect.models.bist.eval.conllu.conll17_ud_eval import run_conllu_eval


# Things that were changed from the original:
# - Removed ConllEntry class, normalize()
# - Changed read_conll() and write_conll() input from file to path
# - Added run_eval(), get_options_dict() and is_conllu()
# - Reformatted code and variable names to conform with PEP8
# - Added legal header


def vocab(conll_path):
    # pylint: disable=missing-docstring
    words_count = Counter()
    pos_count = Counter()
    rel_count = Counter()

    for sentence in read_conll(conll_path):
        words_count.update([node.norm for node in sentence if isinstance(node, ConllEntry)])
        pos_count.update([node.pos for node in sentence if isinstance(node, ConllEntry)])
        rel_count.update([node.relation for node in sentence if isinstance(node, ConllEntry)])

    return words_count, {w: i for i, w in enumerate(words_count.keys())}, list(
        pos_count.keys()), list(rel_count.keys())


def read_conll(path):
    """Yields CoNLL sentences read from CoNLL formatted file.."""
    with open(path, 'r') as conll_fp:
        root = ConllEntry(0, '*root*', '*root*', 'ROOT-POS', 'ROOT-CPOS', '_',
                          -1, 'rroot', '_', '_')
        tokens = [root]
        for line in conll_fp:
            stripped_line = line.strip()
            tok = stripped_line.split('\t')
            if not tok or line.strip() == '':
                if len(tokens) > 1:
                    yield tokens
                tokens = [root]
            else:
                if line[0] == '#' or '-' in tok[0] or '.' in tok[0]:
                    # noinspection PyTypeChecker
                    tokens.append(stripped_line)
                else:
                    tokens.append(
                        ConllEntry(int(tok[0]), tok[1], tok[2], tok[4], tok[3],
                                   tok[5],
                                   int(tok[6]) if tok[6] != '_' else -1,
                                   tok[7], tok[8], tok[9]))
        if len(tokens) > 1:
            yield tokens


def write_conll(path, conll_gen):
    """Writes CoNLL sentences to CoNLL formatted file."""
    with open(path, 'w') as file:
        for sentence in conll_gen:
            for entry in sentence[1:]:
                file.write(str(entry) + '\n')
            file.write('\n')


def run_eval(gold, test):
    """Evaluates a set of predictions using the appropriate script."""
    if is_conllu(gold):
        run_conllu_eval(gold_file=gold, test_file=test)
    else:
        eval_script = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'eval', 'eval.pl')
        with open(test[:test.rindex('.')] + '_eval.txt', 'w') as out_file:
            subprocess.run(['perl', eval_script, '-g', gold, '-s', test], stdout=out_file)


def is_conllu(path):
    """Determines if the file is in CoNLL-U format."""
    return os.path.splitext(path.lower())[1] == '.conllu'


def get_options_dict(activation, lstm_dims, lstm_layers, pos_dims):
    """Generates dictionary with all parser options."""
    return {'activation': activation, 'lstm_dims': lstm_dims, 'lstm_layers': lstm_layers,
            'pembedding_dims': pos_dims, 'wembedding_dims': 100, 'rembedding_dims': 25,
            'hidden_units': 100, 'hidden2_units': 0, 'learning_rate': 0.1, 'blstmFlag': True,
            'labelsFlag': True, 'bibiFlag': True, 'costaugFlag': True, 'seed': 0, 'mem': 0}



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from enum import Enum


class EmbeddingMethod(Enum):
    GLOVE = 'glove'
    GLOVE_OFFLINE = 'glove_offline'
    ELMO = 'elmo'
    ELMO_OFFLINE = 'elmo_offline'


class WikipediaSearchMethod(Enum):
    ONLINE = 'online'
    OFFLINE = 'offline'
    ELASTIC = 'elastic'


class OnlineOROfflineMethod(Enum):
    ONLINE = 'online'
    OFFLINE = 'offline'


class RelationType(Enum):
    NO_RELATION_FOUND = 0
    WIKIPEDIA_REDIRECT_LINK = 1
    WIKIPEDIA_ALIASES = 2
    WIKIPEDIA_DISAMBIGUATION = 3
    WIKIPEDIA_PART_OF_SAME_NAME = 4
    WIKIPEDIA_CATEGORY = 5
    WIKIPEDIA_TITLE_PARENTHESIS = 6
    WIKIPEDIA_BE_COMP = 7
    EXACT_STRING = 8
    FUZZY_FIT = 9
    FUZZY_HEAD_FIT = 10
    SAME_HEAD_LEMMA = 11
    VERBOCEAN_MATCH = 13
    WORDNET_DERIVATIONALLY = 14
    WORDNET_PARTIAL_SYNSET_MATCH = 15
    WORDNET_SAME_SYNSET = 17
    REFERENT_DICT = 18
    WORD_EMBEDDING_MATCH = 19
    WITHIN_DOC_COREF = 20
    OTHER = 21



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import math
from typing import List, Set

from scipy.spatial.distance import cosine as cos

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.embedding.embed_elmo import ElmoEmbedding, \
    ElmoEmbeddingOffline
from nlp_architect.data.cdc_resources.embedding.embed_glove import GloveEmbedding, \
    GloveEmbeddingOffline
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import EmbeddingMethod, \
    RelationType
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


class WordEmbeddingRelationExtraction(RelationExtraction):
    def __init__(self, method: EmbeddingMethod = EmbeddingMethod.GLOVE,
                 glove_file: str = None, elmo_file: str = None, cos_accepted_dist: float = 0.7):
        """
        Extract Relation between two mentions according to Word Embedding cosine distance

        Args:
            method (optional): EmbeddingMethod.{GLOVE/GLOVE_OFFLINE/ELMO/ELMO_OFFLINE}
                (default = GLOVE)
            glove_file (required on GLOVE/GLOVE_OFFLINE mode): str Location of Glove file
            elmo_file (required on ELMO_OFFLINE mode): str Location of Elmo file
        """
        if method == EmbeddingMethod.GLOVE:
            self.embedding = GloveEmbedding(glove_file)
            self.contextual = False
        elif method == EmbeddingMethod.GLOVE_OFFLINE:
            self.embedding = GloveEmbeddingOffline(glove_file)
            self.contextual = False
        elif method == EmbeddingMethod.ELMO:
            self.embedding = ElmoEmbedding()
            self.contextual = True
        elif method == EmbeddingMethod.ELMO_OFFLINE:
            self.embedding = ElmoEmbeddingOffline(elmo_file)
            self.contextual = True

        self.accepted_dist = cos_accepted_dist
        super(WordEmbeddingRelationExtraction, self).__init__()

    def extract_all_relations(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> Set[RelationType]:
        ret_ = set()
        ret_.add(self.extract_sub_relations(mention_x, mention_y,
                                            RelationType.WORD_EMBEDDING_MATCH))
        return ret_

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        if relation is not RelationType.WORD_EMBEDDING_MATCH:
            return RelationType.NO_RELATION_FOUND

        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str
        if StringUtils.is_pronoun(mention_x_str.lower()) or \
                StringUtils.is_pronoun(mention_y_str.lower()):
            if not self.contextual:
                return RelationType.NO_RELATION_FOUND

            if mention_x.mention_context is None or mention_y.mention_context is None:
                return RelationType.NO_RELATION_FOUND

        if self.is_word_embed_match(mention_x, mention_y):
            return RelationType.WORD_EMBEDDING_MATCH

        return RelationType.NO_RELATION_FOUND

    def is_word_embed_match(self, mention_x: MentionDataLight, mention_y: MentionDataLight):
        """
        Check if input mentions Word Embedding cosine distance below above 0.65

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            bool
        """
        match_result = False
        x_embed = self.embedding.get_head_feature_vector(mention_x)
        y_embed = self.embedding.get_head_feature_vector(mention_y)
        # make sure words are not 'unk/None/0'
        if x_embed is not None and y_embed is not None:
            dist = cos(x_embed, y_embed)
            if not math.isnan(dist):
                sim = 1 - dist
                if sim >= self.accepted_dist:
                    match_result = True

        return match_result

    @staticmethod
    def get_supported_relations() -> List[RelationType]:
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.WORD_EMBEDDING_MATCH]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import socket
import pickle
import logging
import sys
import re
from os.path import dirname, join

from bokeh.layouts import column, layout
from bokeh.models import ColumnDataSource, Div, Row, CustomJS
from bokeh.models.widgets import Button, DataTable, TableColumn, CheckboxGroup, MultiSelect
from bokeh.models.widgets.inputs import TextInput
from bokeh.io import curdoc

import nlp_architect.solutions.set_expansion.ui.settings as settings

# pylint: skip-file
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)

vocab = None
vocab_dict = {}
cut_vocab_dict = {}
max_visible_phrases = 5000
working_text = 'please wait...'
fetching_text = 'Fetching vocabulary from server (one time only), this can take few minutes...'
seed_check_text = ''
all_selected_phrases = []
search_flag = False
max_phrase_length = 40
clear_flag = False
expand_columns = [
    TableColumn(field="res", title="Results"),
    TableColumn(field="score", title="Score")
]
empty_table = {'res': 15 * [''], 'score': 15 * ['']}
checkbox_label = "Show extracted term groups" if settings.grouping else "Show extracted phrases"

# create ui components

seed_input_title = 'Please enter a comma separated seed list of terms:'
seed_input_box = TextInput(
    title=seed_input_title, value="", width=450, css_classes=["seed-input"])
annotation_input = TextInput(title="Please enter text to annotate:", value="", width=400,
                             height=80, css_classes=["annotation-input"])
annotation_output = Div(text='', height=30, width=500, style={'padding-left': '35px'})
annotate_button = Button(label="Annotate", button_type="success", width=150,
                         css_classes=["annotation-button"])
group_info_box = Div(text='', height=30, css_classes=["group-div"])
search_input_box = TextInput(title="Search:", value="", width=300)
expand_button = Button(label="Expand", button_type="success", width=150,
                       css_classes=["expand-button"])
clear_seed_button = Button(
    label="Clear", button_type="success", css_classes=['clear_button'], width=50)
export_button = Button(
    label="Export", button_type="success", css_classes=['export_button'], width=100)
expand_table_source = ColumnDataSource(data=empty_table)
expand_table = DataTable(
    source=expand_table_source, columns=expand_columns, width=500, css_classes=['expand_table'])
phrases_list = MultiSelect(
    title="", value=[], options=[], width=300, size=27, css_classes=['phrases_list'])
checkbox_group = CheckboxGroup(
    labels=["Text annotation", checkbox_label], active=[], width=400,
    css_classes=['checkbox_group'])
annotate_checkbox = CheckboxGroup(
    labels=["Text annotation"], active=[], width=400, css_classes=['annotate_checkbox'])
search_box_area = column(children=[Div(height=10, width=200)])
working_label = Div(
    text="", style={'color': 'blue', 'font-size': '15px'})
search_working_label = Div(
    text="", style={'color': 'blue', 'padding-bottom': '0px', 'font-size': '15px'})
seed_check_label = Div(
    text='', style={'font-size': '15px'}, height=20, width=500)
table_layout = Row(
    expand_table)
table_area = column(children=[table_layout])
seed_layout = column(Row(seed_input_box, column(Div(height=14, width=0), clear_seed_button)),
                     expand_button, table_area)
annotation_layout = column(children=[])

phrases_area = column(children=[search_working_label, Div(width=300)])
checkbox_layout = column(children=[checkbox_group, phrases_area])
grid = layout(
    [
        [working_label, Div(width=250), Div(text="<h1>Set Expansion Demo</h1>")],
        [checkbox_layout, seed_layout, Div(width=50),
         column(Div(height=0, width=0), annotation_layout)],
        [group_info_box, Div(width=500), export_button]
    ]
)


# define callbacks

def get_vocab():
    """
    Get vocabulary of the np2vec model from the server
    """
    global vocab
    logger.info('sending get_vocab request to server...')
    received = send_request_to_server(['get_vocab'])
    vocab = received
    for p in vocab:
        if len(p) < max_phrase_length:
            vocab_dict[p] = p
            cut_vocab_dict[p] = p
        else:
            vocab_dict[p] = p[:max_phrase_length - 1] + '...'
            cut_vocab_dict[p[:max_phrase_length - 1] + '...'] = p


def send_request_to_server(request):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        # Connect to server and send data
        sock.connect((settings.expand_host, settings.expand_port))
        logger.info('sending request')
        req_packet = pickle.dumps(request)
        # sock.sendall(bytes(request + "\n", "utf-8"))
        sock.sendall(req_packet)
        # Receive data from the server and shut down
        data = b""
        ctr = 0
        while True:
            packet = sock.recv(134217728)
            logger.info("%s. received: %s", str(ctr), str(len(packet)))
            ctr += 1
            if not packet:
                break
            data += packet
            logger.info('got response, uncompressing')
        received = pickle.loads(data)
        # logger.info("Received: {}".format(received))
        return received
    except EOFError:
        logger.info('No data received')
    finally:
        sock.close()


def row_selected_callback(indices, old, new):
    logger.info('row selected callback')
    global clear_flag, all_selected_phrases
    if not clear_flag and expand_table_source.data != empty_table:
        logger.info('row selected callback. old indices=%s. new indices=%s',
                    str(old), str(new))
        # sync phrases lists:
        old_phrases = [expand_table_source.data['res'][p] for p in old]
        new_phrases = [expand_table_source.data['res'][p] for p in new]
        logger.info('selected_expand was updated: old=%s ,new=%s', str(
            old_phrases), str(new_phrases))
        # phrase was de-selected from expand list:
        for o in old_phrases:
            if o not in new_phrases and \
                    (vocab is not None and vocab_dict[o] in phrases_list.value):
                logger.info('removing %s from vocab selected', o)
                phrases_list.value.remove(vocab_dict[o])
                break
        # new phrase was selected from expand list:
        for n in new_phrases:
            if n not in old_phrases and \
                    (vocab is not None and vocab_dict[n] in
                     phrases_list.options and vocab_dict[n] not in phrases_list.value):
                phrases_list.value.append(vocab_dict[n])
                break
        update_all_selected_phrases()
        seed_input_box.value = get_selected_phrases_for_seed()


def update_all_selected_phrases():
    """
    Sync selected values from both the expand-table and the vocabulary list
    """
    logger.info('update selected phrases')
    global all_selected_phrases
    updated_selected_phrases = all_selected_phrases[:]
    selected_expand = [expand_table_source.data['res'][i] for
                       i in expand_table_source.selected.indices if
                       expand_table_source.data['res'][i] != '']
    selected_vocab = phrases_list.value
    logger.info('selected expand= %s', str(selected_expand))
    logger.info('selected vocab= %s', str(selected_vocab))
    logger.info('current all_selected_phrases= %s', str(all_selected_phrases))
    for x in all_selected_phrases:
        logger.info('x= %s', x)
        if (x in expand_table_source.data['res'] and x not in selected_expand) or (
                vocab is not None and (vocab_dict[x] in phrases_list.options) and (
                vocab_dict[x] not in selected_vocab)
        ):
            logger.info('removing %s', x)
            updated_selected_phrases.remove(x)
    for e in selected_expand:
        if e not in updated_selected_phrases:
            logger.info('adding %s', e)
            updated_selected_phrases.append(e)
    for v in selected_vocab:
        full_v = cut_vocab_dict[v]
        if full_v not in updated_selected_phrases:
            logger.info('adding %s', full_v)
            updated_selected_phrases.append(full_v)
    all_selected_phrases = updated_selected_phrases[:]
    logger.info('all_selected_phrases list was updated: %s', str(all_selected_phrases))


def checkbox_callback(checked_value):
    global search_box_area, phrases_area
    group_info_box.text = ''
    if 0 in checked_value:
        annotation_layout.children = [annotation_input, annotate_button,
                                      annotation_output]
    else:
        annotation_layout.children = []
        annotation_output.text = ""
    if 1 in checked_value:
        if vocab is None:
            working_label.text = fetching_text
            get_vocab()
        if not phrases_list.options:
            working_label.text = working_text
            phrases_list.options = list(
                cut_vocab_dict.keys())[0:max_visible_phrases]  # show the cut representation
        # search_box_area.children = [search_input_box]
        phrases_area.children = [search_input_box, search_working_label, phrases_list]
        working_label.text = ''
    else:
        # search_box_area.children = []
        phrases_area.children = []
        group_info_box.text = ""


def get_expand_results_callback():
    """
    Send to the server the seed to expand and set the results in the expand
    table.
    """
    logger.info('### new expand request')
    working_label.text = working_text
    global seed_check_text, table_area
    try:
        seed_check_label.text = ''
        table_area.children = [table_layout]
        seed = seed_input_box.value
        logger.info('input seed: %s', seed)
        if seed == '':
            expand_table_source.data = empty_table
            return
        seed_words = [x.strip() for x in seed.split(',')]
        bad_words = ''
        for w in seed_words:
            res = send_request_to_server(['in_vocab', w])
            if res is False:
                bad_words += ("'" + w + "',")
        if bad_words != '':
            seed_check_label.text = 'the words: <span class="bad-word">' \
                                    + bad_words[:-1] \
                                    + '</span> are not in the vocabulary and will be ignored'
            logger.info('setting table area')
            table_area.children = [seed_check_label, table_layout]
        logger.info('sending expand request to server with seed= %s', seed)
        received = send_request_to_server(['expand', seed])
        if received is not None:
            res = [x[0] for x in received]
            scores = ["{0:.5f}".format(y[1]) for y in received]
            logger.info('setting table data')
            expand_table_source.data = {
                'res': res,
                'score': scores
            }
        else:
            logger.info('Nothing received from server')
    except Exception as e:
        logger.info('Exception: %s', str(e))
    finally:
        working_label.text = ''


def search_callback(value, old, new):
    group_info_box.text = ''
    search_working_label.text = working_text
    logger.info('search vocab')
    global vocab, phrases_list, all_selected_phrases, search_flag
    search_flag = True
    phrases_list.value = []
    if new == '':
        new_phrases = list(cut_vocab_dict.keys())
    else:
        new_phrases = []
        for x in vocab:
            if x.lower().startswith(new.lower()) and vocab_dict[x] not in new_phrases:
                new_phrases.append(vocab_dict[x])
    phrases_list.options = new_phrases[0:max_visible_phrases]
    if new != '':
        phrases_list.options.sort()
    phrases_list.value = [
        vocab_dict[x] for x in all_selected_phrases if vocab_dict[x] in phrases_list.options]
    logger.info('selected vocab after search= %s', str(phrases_list.value))
    search_working_label.text = ''
    search_flag = False


def vocab_phrase_selected_callback(attr, old_selected, new_selected):
    logger.info('vocab selected')
    if settings.grouping:
        # show group info
        if len(new_selected) == 1:
            res = send_request_to_server(['get_group', new_selected[0]])
            if res is not None:
                group_info_box.text = str(res)
    global clear_flag
    if not clear_flag:
        global all_selected_phrases, search_flag
        if (search_flag):
            return
        logger.info('selected_vocab was updated: old= %s, new= %s', str(
            old_selected), str(new_selected))
        # sync expand table:
        # phrase was de-selected from vocab list:
        expand_selected = [expand_table_source.data['res'][p] for
                           p in expand_table_source.selected.indices]
        for o in old_selected:
            full_o = cut_vocab_dict[o]
            if o not in new_selected and full_o in expand_selected:
                logger.info('%s removed from vocab selected and exists in expand selected', full_o)
                logger.info('removing %s from expand selected indices. index=%s',
                            full_o, str(expand_table_source.data['res'].index(full_o)))
                logger.info('current expand indices: %s',
                            str(expand_table_source.selected.indices))
                expand_table_source.selected.indices.remove(
                    expand_table_source.data['res'].index(full_o))
                logger.info('new expand indices: %s', str(expand_table_source.selected.indices))
                break
        # new phrase was selected from vocab list:
        for n in new_selected:
            full_n = cut_vocab_dict[n]
            logger.info('selected phrase=' + n + ', full phrase=' + full_n)
            if n not in old_selected and full_n in \
                    expand_table_source.data['res'] and full_n not in expand_selected:
                expand_table_source.selected.indices.append(
                    expand_table_source.data['res'].index(full_n))
                break
        update_all_selected_phrases()
        seed_input_box.value = get_selected_phrases_for_seed()


def clear_seed_callback():
    logger.info('clear')
    global all_selected_phrases, table_area, clear_flag
    # table_area.children = []  # needed for refreshing the selections
    clear_flag = True
    seed_input_box.value = ''
    seed_check_label.text = ''
    expand_table_source.selected.indices = []
    phrases_list.value = []
    all_selected_phrases = []
    table_area.children = [table_layout]
    clear_flag = False


def get_selected_phrases_for_seed():
    """
     create the seed string to send to the server
    """
    global all_selected_phrases
    phrases = ''
    for x in all_selected_phrases:
        phrases += x + ', '
    phrases = phrases[:-2]
    return phrases


def expand_data_changed_callback(data, old, new):
    """
    remove the selected indices when table is empty
    """
    if old == empty_table:
        expand_table_source.selected.indices = []


def annotate_callback():
    try:
        annotation_output.text = working_text
        user_text = annotation_input.value
        # if len(user_text) == 0 :
        #    annotation_output.text = "Please provaide valid text to annotate"
        if len(seed_input_box.value) == 0:
            out_text = "No seed to compare to"
        else:
            out_text = user_text
            seed = [x.strip() for x in seed_input_box.value.split(',')]
            res = send_request_to_server(['annotate', seed, user_text])
            logger.info("res:%s", str(res))
            if len(res) == 0:
                out_text = "No results found"
            for np in res:
                pattern = re.compile(r'\b' + np + r'\b')
                out_text = re.sub(pattern, mark_phrase_tag(np), out_text)
        annotation_output.text = out_text
    except Exception as e:
        annotation_output.text = "An error occured"
        logger.error("Error: %s", e)


def mark_phrase_tag(text):
    return '<phrase>' + text + '</phrase>'


# set callbacks

expand_button.on_click(get_expand_results_callback)
expand_table_source.selected.on_change('indices', row_selected_callback)
expand_table_source.on_change('data', expand_data_changed_callback)
checkbox_group.on_click(checkbox_callback)
search_input_box.on_change('value', search_callback)
phrases_list.on_change('value', vocab_phrase_selected_callback)
clear_seed_button.on_click(clear_seed_callback)
with open(join(dirname(__file__), "download.js")) as f:
    code = f.read()
export_button.callback = CustomJS(args=dict(source=expand_table_source),
                                  code=code)
annotate_button.on_click(annotate_callback)
# table_area.on_change('children', table_area_change_callback)

# arrange components in page

doc = curdoc()
main_title = "Set Expansion Demo"
doc.title = main_title
doc.add_root(grid)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import csv
import logging
import operator
import os
import sys
from os import path
from pathlib import Path
from shutil import copyfile

import numpy
import pandas as pd
from fastText import load_model
from sklearn.manifold import TSNE

from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.io import check_size, validate_existing_filepath
from nlp_architect.utils.text import simple_normalizer

dir = str(LIBRARY_OUT / 'trend-analysis-data')
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)
target_topics_path = path.join(dir, 'target_topics.csv')
ref_topics_path = path.join(dir, 'ref_topics.csv')


def analyze(target_data, ref_data, tar_header, ref_header, top_n=10000, top_n_vectors=500,
            re_analysis=False, tfidf_w=0.5, cval_w=0.5, lm_w=0):
    """
    Compare a topics list of a target data to a topics list of a reference data
    and extract hot topics, trends and clusters. Topic lists can be generated
    by running topic_extraction.py

    Args:
        target_data: A list of topics with importance scores extracted from the tagret corpus
        ref_data: A list of topics with importance scores extracted from the reference corpus
        tar_header: The header to appear for the target topics graphs
        ref_header: The header to appear for the reference topics graphs
        top_n (int): Limit the analysis to only the top N phrases of each list
        top_n_vectors (int): The number of vectors to include in the scatter
        re_analysis (Boolean): whether a first analysis has already been made or not
        tfidf_w (Float): the TF_IDF weight for the final score calculation
        cval_w (Float): the C_Value weight for the final score calculation
        lm_w (Float): the Language-Model weight for the final score calculation
    """
    hash2group = {}
    rep2rank = {}  # the merged list of groups extracted from both corpora.
    #  Will be sorted by their rank.
    in_model_count = 0
    create_clusters = False
    try:
        if not re_analysis:  # first analysis, not through ui
            copyfile(target_data, target_topics_path)
            copyfile(ref_data, ref_topics_path)
        calc_scores(target_data, tfidf_w, cval_w, lm_w, target_topics_path)
        calc_scores(ref_data, tfidf_w, cval_w, lm_w, ref_topics_path)
        # unify all topics:
        with open(ref_data) as f:
            topics1 = sum(1 for _ in f)
        with open(target_data) as f:
            topics2 = sum(1 for _ in f)
        sum_topics = topics1 + topics2
        logger.info("sum of all topics= %s", str(sum_topics))
        merge_phrases(ref_topics_path, True, hash2group, rep2rank, top_n, sum_topics)
        merge_phrases(target_topics_path, False, hash2group, rep2rank, top_n, sum_topics)
        logger.info("Total number of evaluated topics: %s", str(len(rep2rank)))

        # compute 2D space clusters if model exists:
        w2v_loc = path.join(dir, 'W2V_Models/model.bin')
        if os.path.isfile(w2v_loc):
            all_topics_sorted = sorted(rep2rank, key=rep2rank.get)
            top_n_scatter = len(
                all_topics_sorted) if top_n_vectors is None else top_n_vectors
            scatter_group = all_topics_sorted[0:top_n_scatter]
            np_scat, x_scat, y_scat, in_model_count =\
                compute_scatter_subwords(scatter_group, w2v_loc)
            if np_scat is not None and x_scat is not None:
                create_clusters = True
                for j in range(len(np_scat)):
                    hash2group[simple_normalizer(np_scat[j])] += (x_scat[j],
                                                                  y_scat[j])
        # prepare reports data:
        groups_r = list(
            filter(lambda x: x[2] == 0 or x[2] == 2, hash2group.values()))
        groups_t = list(
            filter(lambda x: x[2] == 1 or x[2] == 2, hash2group.values()))
        trends = list(
            filter(lambda x: x[2] == 2, hash2group.values()))  # all trends
        groups_r_sorted = sorted(
            groups_r, key=operator.itemgetter(1), reverse=True)
        groups_t_sorted = sorted(
            groups_t, key=operator.itemgetter(3), reverse=True)
        trends_sorted = sorted(
            trends, key=operator.itemgetter(6), reverse=True)  # sort by t_score

        # save results:
        save_report_data(hash2group, groups_r_sorted, groups_t_sorted,
                         trends_sorted, all_topics_sorted, create_clusters,
                         tar_header, ref_header, tfidf_w, cval_w, lm_w,
                         in_model_count, top_n_scatter)
        logger.info('Done analysis.')
    except Exception as e:
        logger.error(str(e))


def save_report_data(hash2group, groups_r_sorted, groups_t_sorted,
                     trends_sorted, all_topics_sorted, create_clusters,
                     target_header, ref_header, tfidf_w, cval_w, freq_w,
                     in_model_count, top_n_scatter):
    filter_output = path.join(dir, 'filter_phrases.csv')
    data_output = path.join(dir, 'graph_data.csv')
    logger.info('writing results to: %s and: %s ', str(data_output), str(filter_output))

    reports_column = [
        'Top Topics (' + ref_header + ')',
        'Top Topics (' + target_header + ')',
        'Hot Trends',
        'Cold Trends',
        'Custom Trends'
    ]
    if create_clusters:
        reports_column.extend([
            'Trend Clustering',
            'Topic Clustering (' + ref_header + ')',
            'Topic Clustering (' + target_header + ')'
        ])
    headers = [
        'reports',
        'ref_topic', 'ref_imp', 'x_ref', 'y_ref',
        'tar_topic', 'tar_imp', 'x_tar', 'y_tar',
        'trends', 'change', 't_score', 'x_tre', 'y_tre', 'weights',
        'in_w2v_model_count', 'top_n_scatter'
    ]
    filter_headers = [
        'topics',
        'valid',
        'custom'
    ]
    weights = [tfidf_w, cval_w, freq_w]

    with open(data_output, 'wt', encoding='utf-8') as data_file:
        with open(filter_output, 'wt', encoding='utf-8') as filter_file:
            data_writer = csv.writer(data_file, delimiter=',')
            filter_writer = csv.writer(filter_file, delimiter=',')
            data_writer.writerow(headers)
            filter_writer.writerow(filter_headers)
            for i in range(len(hash2group.keys())):
                try:
                    new_row = ()
                    new_row += (reports_column[i],) if i < len(
                        reports_column) else ('',)
                    if i < len(groups_r_sorted):
                        new_row += (
                            groups_r_sorted[i][0], groups_r_sorted[i][1])
                        # 'only b' type tuple
                        if groups_r_sorted[i][2] == 0:
                            new_row += (
                                groups_r_sorted[i][-2],
                                groups_r_sorted[i][-1]) if len(
                                groups_r_sorted[i]) > 4 else (-1, -1)
                        else:  # 'trend' type tuple
                            new_row += (
                                groups_r_sorted[i][-2],
                                groups_r_sorted[i][-1]) if len(
                                groups_r_sorted[i]) > 7 else (-1, -1)
                    else:
                        new_row += ('', '', '', '')
                    if len(groups_t_sorted) > i:
                        new_row += (
                            groups_t_sorted[i][0], groups_t_sorted[i][3])
                        # 'only a' type tuple
                        if groups_t_sorted[i][2] == 1:
                            new_row += (groups_t_sorted[i][-2],
                                        groups_t_sorted[i][-1]) if len(
                                groups_t_sorted[i]) > 5 else (-1, -1)
                        else:
                            new_row += (groups_t_sorted[i][-2],
                                        groups_t_sorted[i][-1]) if len(
                                groups_t_sorted[i]) > 7 else (-1, -1)
                    else:
                        new_row += ('', '', '', '')
                    if len(trends_sorted) > i:
                        new_row += (
                            trends_sorted[i][0], trends_sorted[i][4], trends_sorted[i][6])
                        new_row += (
                            trends_sorted[i][-2],
                            trends_sorted[i][-1]) if len(
                            trends_sorted[i]) > 7 else (-1, -1)
                    else:
                        new_row += ('', '', '', '')
                    new_row += (weights[i],) if i < len(
                        weights) else ('',)
                    new_row += (in_model_count, top_n_scatter) if i == 0 else ('',)
                    data_writer.writerow(new_row)

                    filter_row = (all_topics_sorted[i], 1, 0)
                    filter_writer.writerow(filter_row)
                except Exception as e:
                    logger.error(
                        "Error while writing analysis results. "
                        "iteration #: %s. Error: %s", str(i), str(e))
            if len(hash2group.keys()) < len(reports_column):
                for i in range(len(hash2group.keys()), len(reports_column)):
                    new_row = ()
                    new_row += (reports_column[i], '', '', '', '', '', '', '',
                                '', '', '', '', '', '', '', '', '')
                    data_writer.writerow(new_row)


def compute_scatter_subwords(top_groups, w2v_loc):
    """
    Compute 2D vectors of the provided phrases groups

    Args:
        top_groups: A list of group-representative phrases (List(String))
        w2v_loc: A path to a w2v model (String)

    Returns:
        A tuple (phrases, x, y, n)
        WHERE:
        phrases: A list of phrases that are part of the model
        x: A DataFrame column as the x values of the phrase vector
        y: A DataFrame column as the y values of the phrase vector
        n: The number of computed vectors
    """
    x_feature, np_indices = [], []
    in_ctr = 0

    try:
        if path.isfile(w2v_loc):
            logger.info("computing scatter on %s groups", str(len(top_groups)))
            model = load_model(w2v_loc)
            for a in top_groups:
                x_feature.append(numpy.array(model.get_word_vector(a)))
                np_indices.append(a)
                in_ctr += 1
            logger.info("topics found in model out of %s: %s.", str(len(top_groups)), str(in_ctr))
            if len(np_indices) == 0:
                logger.error('no vectors extracted')
                return None, None, None, 0
            logger.info('computing TSNE embedding...')
            tsne = TSNE(n_components=2, random_state=0, method='exact')
            numpy.set_printoptions(suppress=True)
            x_tsne = tsne.fit_transform(x_feature)
            df = pd.DataFrame(
                x_tsne, index=np_indices, columns=['x', 'y'])
            return np_indices, df['x'], df['y'], in_ctr
        logger.error('no model found for cumputing scatter, skipping step.')
        return None, None, None, 0
    except Exception as e:
        logger.error(str(e))


def calc_scores(scores_file, tfidf_w, cval_w, lm_w, output_path):
    """
    Given a topic list with tf_idf,c_value,language_model scores, compute
     a final score for each phrases group according to the given weights

    Args:
        scores_file (String): A path to the file with groups and raw scores
        tfidf_w (Float): the TF_IDF weight for the final score calculation
        cval_w (Float): the C_Value weight for the final score calculation
        lm_w (Float): the Language-Model weight for the final score calculation
        output_path: A path for the output file of final scores (String)
    """
    logger.info("calculating scores for file: %s with: tfidf_w=%s, cval_w=%s,"
                " freq_w=%s", scores_file, str(tfidf_w), str(cval_w), str(lm_w))
    with open(scores_file, encoding='utf-8', errors='ignore') as csv_file:
        lines = csv.reader(csv_file, delimiter=',')
        final_list = []
        for group, tfidf, cval, freq in lines:
            score = (float(tfidf) * tfidf_w) + (float(cval) * cval_w) + (float(freq) * lm_w)
            final_list.append((group, score))
        sorted_list = sorted(final_list, key=lambda tup: tup[1], reverse=True)
    save_scores_list(sorted_list, output_path)


def save_scores_list(scores, file_path):
    """
    Save the list of topics-scores into a file

    Args:
        scores: A list of topics (groups) with scores
        file_path: The output file path
    """
    logger.info('saving np extraction results to: %s', file_path)
    with open(file_path, 'wt', encoding='utf-8', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=',')
        ctr = 0
        for topics, imp in scores:
            try:
                row = (topics, str(imp))
                writer.writerow(row)
                ctr += 1
            except Exception as e:
                logger.error("Error while writing scores to file. iteration #: %s. Error: %s", str(
                    ctr), str(e))


def merge_phrases(data, is_ref_data, hash2group, rep2rank, top_n, topics_count):
    """
    Analyze the provided topics data and detect trends (changes in importance)

    Args:
        data: A list of topics with importance scores
        is_ref_data (Boolean): Was the data extracted from the target/reference corpus
        hash2group: A dictionary storing the data of each topic
        rep2rank: A dict of all groups representatives and their ranks
        top_n (int): Limit the analysis to only the top N phrases of each list
        topics_count (int): The total sum of all topics extracted from both corpora
    """
    logger.info('merge and compare groups for data: %s', str(data))
    ctr = 0
    if not Path(data).exists():
        logger.error('invalid csv file: %s', str(data))
        sys.exit()
    try:
        with open(data, encoding='utf-8', errors='ignore') as csv_file:
            topics = csv.reader(csv_file, delimiter=',')
            for group, imp in topics:
                if ctr == top_n:
                    break
                try:
                    rep = clean_group(group).strip()
                    imp = float(imp) * 100.0
                    rank = ctr + 1
                    hash_id = simple_normalizer(rep)
                    if hash_id not in hash2group:
                        rep2rank[rep] = rank
                        if is_ref_data:
                            hash2group[hash_id] = (rep, imp, 0, rank)
                        else:
                            hash2group[hash_id] = (rep, 0, 1, imp, rank)
                    elif not is_ref_data:
                        data_b = hash2group[hash_id]
                        if data_b[2] == 0:  # create a trend only in comparison to the
                            #  ref topics, ignore cases of different topics have the
                            # same hash or same topic was extracted twice from the
                            # same data
                            old_rep = data_b[0]
                            old_rank = data_b[3]
                            rep2rank[old_rep] = int((rank + old_rank) / 2)  # rank
                            #  of topic that appear in both corpora calculated as
                            #  the avarage of ranks
                            change = float(imp) - float(data_b[1])
                            t_score = (topics_count - (old_rank + rank)) * abs(change)
                            hash2group[hash_id] = (
                                old_rep, float(data_b[1]), 2, imp, change,
                                abs(change), t_score)  # trend phrase
                    ctr += 1
                except Exception as e:
                    logger.error('bad line: %s. Error: %s', str(ctr), str(e))
    except Exception as e:
        logger.error('Error: %s. Is %s a valid csv file?', str(e), str(data))
        sys.exit()


def clean_group(phrase_group):
    """
   Returns the shortest element in a group of phrases

    Args:
        phrase_group (String): a group of phrases separated by ';'

    Returns:
        The shortest phrase in the group (String)
    """
    text = [x.lstrip() for x in phrase_group.split(';')]
    return min(text, key=len)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='trend_analysis.py')
    parser.add_argument('target_topics', metavar='target_topics', type=validate_existing_filepath,
                        help='a path to a csv topic-list extracted from the '
                             'target corpus')
    parser.add_argument('ref_topics', metavar='ref_topics', type=validate_existing_filepath,
                        help='a path to a csv topic-list extracted from the '
                             'reference corpus')
    parser.add_argument('--top_n', type=int, action=check_size(0, 100000), default=10000,
                        help='compare only top N topics (default: 10000)')
    parser.add_argument('--top_vectors', type=int, action=check_size(0, 100000), default=500,
                        help='include only top N vectors in the scatter graph (default: 500)')
    args = parser.parse_args()
    analyze(args.target_topics, args.ref_topics, args.target_topics,
            args.ref_topics, args.top_n, args.top_vectors)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import socketserver
import argparse
import pickle
import logging
import sys
from nlp_architect.solutions.set_expansion import set_expand
from nlp_architect.utils.io import validate_existing_filepath, check_size
from nlp_architect.solutions.set_expansion.prepare_data import load_parser, extract_noun_phrases

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


class MyTCPHandler(socketserver.BaseRequestHandler):
    """
    A simple server to load the w2v model and handle expand requests from the
    ui
    """

    def handle(self):
        logger.info("handling expand request")
        res = ''
        self.data = pickle.loads(self.request.recv(10240))
        logger.info('request data: %s', self.data)
        req = self.data[0]
        if req == 'get_vocab':
            logger.info('getting vocabulary')
            res = se.get_vocab()
        elif req == 'in_vocab':
            term = self.data[1]
            res = se.in_vocab(term)
        elif req == 'get_group':
            term = self.data[1]
            res = se.get_group(term)
        elif req == 'annotate':
            seed = self.data[1]
            text = self.data[2]
            res = self.annotate(text, seed)
            logger.info("res:%s", str(res))
        elif req == 'expand':
            logger.info('expanding')
            data = [x.strip() for x in self.data[1].split(',')]
            res = se.expand(data)
        logger.info('compressing response')
        packet = pickle.dumps(res)
        logger.info('response length= %s', str(len(packet)))
        logger.info('sending response')
        self.request.sendall(packet)
        logger.info('done')

    @staticmethod
    def annotate(text, seed):
        np_list = []
        docs = [text]
        spans = extract_noun_phrases(docs, nlp, args.chunker)
        for x in spans:
            np = x.text
            if np not in np_list:
                np_list.append(np)
        logger.info("np_list=%s", str(np_list))
        return se.similarity(np_list, seed, args.similarity)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='expand_server.py')
    parser.add_argument('model_path', metavar='model_path',
                        type=validate_existing_filepath,
                        help='a path to the w2v model file')
    parser.add_argument('--host', type=str, default='localhost',
                        help='set port for the server', action=check_size(1, 20))
    parser.add_argument('--port', type=int, default=1234,
                        help='set port for the server', action=check_size(0, 65535))
    parser.add_argument('--grouping', action='store_true', default=False, help='grouping mode')
    parser.add_argument('--similarity', default=0.5, type=float,
                        action=check_size(0, 1), help='similarity threshold')
    parser.add_argument('--chunker', type=str, choices=['spacy', 'nlp_arch'],
                        help='spacy chunker or \'nlp_arch\' for NLP Architect NP Extractor')
    args = parser.parse_args()

    port = args.port
    model_path = args.model_path
    logger.info("loading model")
    se = set_expand.SetExpand(model_path, grouping=args.grouping)
    logger.info("loading chunker")
    nlp = load_parser(args.chunker)
    logger.info("loading server")
    HOST, PORT = args.host, port
    server = socketserver.TCPServer((HOST, PORT), MyTCPHandler)
    logger.info("server loaded")
    server.serve_forever()
